/* SAMHAIN file system integrity testing                                   */
/* Copyright (C) 1999 Rainer Wichmann                                      */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */


#include <stdio.h>
#include <pwd.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>

#include "config.h"

#include "samhain.h"
#include "sh_utils.h"
#include "sh_tiger.h"

#ifndef HAVE_URANDOM


typedef struct {
  char   * path;
  char   * command;
  int      pipeFD;
  int      isset;
  FILE   * pipe;
} sourcetable_t;

sourcetable_t source[] = {
  { "/usr/bin/vmstat",
    "/usr/bin/vmstat",
    0,
    0,
    NULL },
  { "/bin/df",
    "/bin/df",
    0,
    0,
    NULL },
  { "/usr/sbin/arp",
    "/usr/sbin/arp -a",
    0,
    0,
    NULL },
  { "/bin/netstat",
    "/bin/netstat -n",
    0,
    0,
    NULL },
  { "/usr/bin/procinfo",
    "/usr/bin/procinfo -a",
    0,
    0,
    NULL },
  { "/usr/bin/fortune",
    "/usr/bin/fortune -a",
    0,
    0,
    NULL },
  { "/bin/ps",
    "/bin/ps auxl",
    0,
    0,
    NULL },
  { "/usr/bin/last",
    "/usr/bin/last",
    0,
    0,
    NULL },
  { "/usr/bin/uptime",
    "/usr/bin/uptime",
    0,
    0,
    NULL },
  { NULL,
    NULL,
    0,
    0,
    NULL }
};

char * sh_entropy()
{
  char * buffer;
  int    i, imax, icount;
  unsigned long select_now = 0;
  int    bufcount;
  struct timeval tv;
  fd_set fds;
  int    maxFD = 0;
  int    count;
  static char   keybuf[KEY_LEN+1];

  struct passwd * tempres;
  gid_t store_egid;
  uid_t store_euid;

  store_egid = getegid();
  store_euid = geteuid();

  /* drop root
   */
  i = 0; 
  if (0 == geteuid()) {  
    /* we are root, so lets drop 'root' 
     */
    tempres = getpwnam(DEFAULT_IDENT);
    if (NULL != tempres) {
      i = setegid(tempres->pw_gid);  
      if (i == 0) i = seteuid(tempres->pw_uid);
    } else {
      i = -1;
    }
  }
  if (i == -1) {
    perror ("sh_entropy - drop root");
    exit (EXIT_FAILURE);
  }

  i = 0; icount = 0;
  buffer = SH_ALLOC(BUFSIZ+2);
  FD_ZERO(&fds);   

  while (source[i].path != NULL) {
    if ( access (source[i].path, X_OK) == 0) {

      /* source exists
       */
      source[i].pipe   = popen  ( source[i].command, "r" );
      source[i].pipeFD = fileno ( source[i].pipe    );

      maxFD = (source[i].pipeFD > maxFD) ? source[i].pipeFD : maxFD;
 
      fcntl( source[i].pipeFD, F_SETFL, O_NONBLOCK);
      FD_SET( source[i].pipeFD, &fds );
      source[i].isset = 1;
      ++icount;
    }
    ++i;
  }

  imax       = i;

  tv.tv_sec  = 1;
  tv.tv_usec = 0;

  bufcount   = 0;

  while ( (icount > 0) && (bufcount < BUFSIZ) ) {

    if ( select (maxFD+1, &fds, NULL, NULL, &tv) == -1) 
      break;

    /* let's not hang on forever
     */
    ++select_now;
    if ( select_now > 60 ) 
      break;
    
    for (i = 0; i < imax; ++i) {

      if ( FD_ISSET (source[i].pipeFD, &fds) ) {
	count = fread (&buffer[bufcount], 
		       1, 
		       BUFSIZ-bufcount, 
		       source[i].pipe );
	if (count == 0) {
	  source[i].isset = 0;
	  FD_CLR (source[i].pipeFD, &fds);
	  pclose ( source[i].pipe );
	  --icount;
	}
	bufcount += count;

      } else {

	/* select() modifies the set, so lets reset it
	 */
	if (source[i].isset == 1) 
	  FD_SET( source[i].pipeFD, &fds );

      }
    }
  }

  for (i = 0; i < imax; ++i) {
    if (source[i].isset == 1)
      pclose ( source[i].pipe );
  }

  buffer[bufcount] = '\0';

  if (bufcount > 0) {
    strlcpy (keybuf, sh_tiger_hash(buffer, TIGER_DATA), KEY_LEN+1); 
  } else {
    keybuf[0] = '\0';
  }

  memset (buffer, 'X', BUFSIZ);
  SH_FREE(buffer);

  seteuid(store_euid);
  setegid(store_egid);

  return keybuf;
}

#endif
