/* SAMHAIN file system integrity testing                                   */
/* Copyright (C) 1999 Rainer Wichmann                                      */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <errno.h>

#include "config.h"
#include "samhain.h"
#include "sh_error.h"
#include "sh_getopt.h"
#include "sh_files.h"
#include "sh_utils.h"
#include "sh_mail.h"



#define HAS_ARG_NO  0
#define HAS_ARG_YES 1
#define DROP_PRIV_NO  0
#define DROP_PRIV_YES 1


typedef struct options {
  const char * longopt;
  const char   shortopt;
  const char * usage;
  int          hasArg;
  int (*func)(char * opt);
} opttable_t;

static int sh_getopt_usage (char * dummy);
static int sh_getopt_copyright (char * dummy);



opttable_t table[] = {
  { "check-file-noignore",  
    'F', 
    "Check a file, report all modifications",  
    HAS_ARG_YES,
    sh_files_pushfile_noig },
  { "check-file-readonly",  
    'R', 
    "Check a file, ignore access time",  
    HAS_ARG_YES, 
    sh_files_pushfile_ro },
  { "check-file-loggrow",  
    'L', 
    "Check a file, ignore timestamp, growing size, signature",  
    HAS_ARG_YES, 
    sh_files_pushfile_glog },
  { "check-dir-readonly",  
    'D', 
    "Check a directory, ignore access time",  
    HAS_ARG_YES, 
    sh_files_pushdir_ro },
  { "recursion",  
    'r', 
    "Set recursion level for directories",  
    HAS_ARG_YES, 
    sh_files_setrecursion },
  { "set-syslog-severity",  
    'S', 
    "Set severity for syslog",  
    HAS_ARG_YES, 
    sh_error_set_syslog },
  { "set-log-severity",  
    's', 
    "Set severity for logfile",  
    HAS_ARG_YES, 
    sh_error_setlog },
  { "set-mail-severity",  
    'm', 
    "Set severity for mail",  
    HAS_ARG_YES, 
    sh_error_setseverity },
  { "set-print-severity",  
    'p', 
    "Set the severity for print log",  
    HAS_ARG_YES, 
    sh_error_setprint },
  { "set-checksum-test",  
    't', 
    "Set checksum testing to init/check",  
    HAS_ARG_YES, 
    sh_util_setchecksum },
  { "set-debug-level",  
    'l', 
    "Set the debug level",  
    HAS_ARG_YES, 
    sh_error_setdebug },
  { "verify-log",  
    'V', 
    "Verify the audit trail",  
    HAS_ARG_YES, 
    sh_error_logverify },
  { "verify-mail",  
    'X', 
    "Verify the mailbox",  
    HAS_ARG_YES, 
    sh_mail_sigverify 
  },
  { "verify-string",  
    'v', 
    "Verify a string (print its hash)",  
    HAS_ARG_YES, 
    sh_error_verify },
  { "deamon",  
    'd', 
    "Start as deamon",  
    HAS_ARG_NO, 
    sh_unix_setdeamon },
  { "copyright",  
    'c', 
    "Print copyright information",  
    HAS_ARG_NO, 
    sh_getopt_copyright },
  { "help",  
    'h', 
    "Print usage information",  
    HAS_ARG_NO, 
    sh_getopt_usage },
  /* last entry -- required !! -- */
  { NULL, 
    '\0',     
    NULL,  
    HAS_ARG_NO, 
    NULL }
};

static int sh_getopt_copyright (char * dummy)
{
  fprintf (stdout, "Copyright (C) 1999 Rainer Wichmann.\n\n");

  fprintf (stdout, 
	   "This program is free software; "\
	   "you can redistribute it and/or modify\n");
  fprintf (stdout, "it under the terms of "\
	   "the GNU General Public License as published by\n");
  fprintf (stdout, "the Free Software Foundation; "\
	   "either version 2 of the License, or\n");
  fprintf (stdout, "(at your option) any later version.\n\n");

  fprintf (stdout, "This program is distributed in the hope "\
	   "that it will be useful,\n");
  fprintf (stdout, "but WITHOUT ANY WARRANTY; "\
	   "without even the implied warranty of\n");
  fprintf (stdout, "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE."\
	   " See the\n");
  fprintf (stdout, "GNU General Public License for more details.\n\n");

  fprintf (stdout, "You should have received a copy of the "\
	   "GNU General Public License\n");
  fprintf (stdout, "along with this program; "\
	   "if not, write to the Free Software\n");
  fprintf (stdout, "Foundation, Inc., 59 Temple Place - Suite 330, "\
	   "Boston, MA  02111-1307, USA.\n");

  if (sh.locked == BAD)
    _exit (EXIT_SUCCESS);
  else
    exit (EXIT_SUCCESS);

}
  
static int sh_getopt_usage (char * dummy)
{
  int  i;
  char fmt[64];
#ifdef MEM_DEBUG
  char opts[64] = { '\0' };
#endif

  fprintf (stdout, "This is %s, version %s, (c) 1999 Rainer Wichmann.\n", 
	   PACKAGE, VERSION);
  fprintf (stdout, "This software comes with ABSOLUTELY NO WARRANTY. ");
  fprintf (stdout, "Use at own risk.\n");

  fprintf (stdout, "Usage:\n\n");

  for (i = 0; table[i].shortopt != '\0'; ++i) {

#ifdef MEM_DEBUG
    if (strchr(opts, table[i].shortopt) != NULL)
      fprintf (stdout, "Short option char collision !");
    opts[i] = table[i].shortopt;
#endif

    if (table[i].hasArg == HAS_ARG_NO) {
      if (strlen(table[i].longopt) < 10) 
	strcpy(fmt, "-%c,        --%-s,\t\t\t %s\n");    /* known to fit  */
      else if (strlen(table[i].longopt) < 17)
	strcpy(fmt, "-%c,        --%-s,\t\t %s\n");      /* known to fit  */
      else 
	strcpy(fmt, "-%c,        --%-s,\t %s\n");        /* known to fit  */
      fprintf (stdout,
	       fmt,
	       table[i].shortopt,
	       table[i].longopt,
	       table[i].usage);
    } else {
      if (strlen(table[i].longopt) < 12) 
	strcpy(fmt, "-%c <arg>,  --%-s=<arg>,\t\t %s\n");/* known to fit  */
      else 
	strcpy(fmt, "-%c <arg>,  --%-s=<arg>,\t %s\n");  /* known to fit  */
      fprintf (stdout,
	       fmt,
	       table[i].shortopt,
	       table[i].longopt,
	       table[i].usage);
      
    }
  }

  fprintf (stdout, "\nReport bugs to rwichmann@hs.uni-hamburg.de.\n");

  fflush(stdout);

  if ( dummy != NULL) {
    if (strcmp( dummy, "fail") == 0 ) {
      if (sh.locked == BAD)
	_exit (EXIT_FAILURE);
      else
	exit (EXIT_FAILURE);
    }  
  } else {
    if (sh.locked == BAD)
      _exit (EXIT_SUCCESS);
    else
      exit (EXIT_SUCCESS);
  }

  if (sh.locked == BAD)
    _exit (EXIT_SUCCESS);
  else
    exit (EXIT_SUCCESS);
}

  

int sh_getopt_get (int argc, char * argv[])
{
  int           count   = 0;
  unsigned long len     = 0;
  int           foundit = 0;
  int           i;
  char        * theequal;

  /* return if no args
   */
  if (argc < 2) return 0;
 
  while (argc > 1  && argv[1][0] == '-') {

    /* initialize
     */
    foundit = 0;
    len     = strlen (argv[1]);
    
    /* a '-' with no argument: error
     */
    if (len == 1) {
      sh_getopt_usage("fail");
    }

    /* a '--' with no argument: stop argument processing
     */
    if (len == 2 && argv[1][1] == '-') {
      return count;
    }
    
    /* a short option: process it
     */
    if (len == 2) {

      for (i = 0; table[i].shortopt != '\0'; ++i) {
      
	if ( table[i].shortopt == argv[1][1] ) {
	  foundit = 1;
	  if ( table[i].hasArg == HAS_ARG_YES ) {
	    if (argc < 3) { 
	      /* argument required, but no avail */
	      fprintf (stderr, "Error: missing argument\n");
	      sh_getopt_usage("fail");
	    } else {
	      /* call function with argument */
	      --argc; ++argv;
	      (* table[i].func )(argv[1]);
	      break;
	    }
	  } else {
	    (* table[i].func )(NULL);
	    break;
	  }
	}

      }

      /* 'break' should get here */
      if (foundit == 1) {
	--argc; ++argv;
	continue;
      } else {
	/* unrecognized short option */
	fprintf (stderr, "Error: unrecognized short option\n");
	sh_getopt_usage("fail");
      }
    }

    /* a long option: process it
     */
    if (len > 2) {

      for (i = 0; table[i].longopt != NULL; ++i) {
      
	if (strncmp(table[i].longopt, 
		    &argv[1][2], 
		    strlen(table[i].longopt)) == 0 ) {
	  foundit = 1; 
	  if ( table[i].hasArg == HAS_ARG_YES ) {
	    if ( (theequal = strchr(argv[1], '=')) == NULL) { 
	      fprintf (stderr, "Error: missing argument\n");
	      /* argument required, but no avail */
	      sh_getopt_usage("fail");
	    } else {
	      if (strlen (theequal) > 1) {
		++theequal;
		/* call function with argument */
		(* table[i].func )(theequal);
		break;
	      } else {
		fprintf (stderr, "Error: invalid argument\n");
		/* argument required, but no avail */
		sh_getopt_usage("fail");
	      }
	    }
	  } else {
	    (* table[i].func )(NULL);
	    break;
	  }
	}

      }

      /* 'break' should get here */
      if (foundit == 1) {
	++count;
	--argc; 
	++argv;
	continue;
      } else {
	/* unrecognized long option */
	fprintf (stderr, "Error: unrecognized long option\n");
	sh_getopt_usage("fail");
      }
    }


  }

  return count;

}
