/* SAMHAIN file system integrity testing                                   */
/* Copyright (C) 1999 Rainer Wichmann                                      */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <signal.h>
#include <setjmp.h>


#include "config.h"

#include "samhain.h"
#include "sh_error.h"
#include "sh_unix.h"
#include "sh_tiger.h"
#include "sh_mail.h"
#include "sh_utils.h"

static int failedMail = GOOD;

typedef struct mail_msg {
  char   msg[BUFSIZ + 53 + 18];
  struct mail_msg * next;
} mail_msg_t;

static mail_msg_t * mailStack = NULL;

/*********************************************
 *  utility function for verifying mails
 *********************************************/

int sh_mail_sigverify (char * s)
{
  int    fd;
  int    i;
  char * buf;
  char * bufc;
  char   key[81];
  char   number[81];
  long   numsig;
  char   key2[KEY_MAX+1];
  char   key3[KEY_MAX+1];

  (void) sh_unix_droppriv (); /* do not honour SUID settings */

  sh_error_logoff();

  if (s == NULL || strlen(s) >= PATH_MAX) 
    exit (EXIT_FAILURE);

  /* open the file, then check it 
   */
  if ( (fd = sh_unix_fileopen (s, 'r', 
			       SH_ISFILE, SH_ERR_FATAL)) == (-1))
    {
      fprintf(stderr, _("Could not open file %s\n"), s);
      _exit (EXIT_FAILURE);
    }

  /* get key
  fprintf(stdout, "Enter key: ");
  fgets(key2, KEY_MAX+1, stdin); 
   */

  buf     = SH_ALLOC( (unsigned long)(SH_BUFSIZE+1));
  bufc    = SH_ALLOC( (unsigned long)(SH_MAXBUF+1));

  while (1)
    {
      buf[0]  = '\0';
      bufc[0] = '\0';

      /* find start of next message
       */
      while (0 != strncmp(buf,   "<--- MESSAGE   ---->",
			  sizeof("<--- MESSAGE   ---->")-1)) 
	{
	  (void) sh_unix_getline (fd, buf, SH_BUFSIZE);
	  if (buf[0] == '\0') 
	    _exit (EXIT_SUCCESS);
	}
      
      /* read message, compress into bufc
       */
      while (1)
	{
	  (void) sh_unix_getline (fd, buf, SH_BUFSIZE);
	  if (0 == strncmp(buf,   "<--- SIGNATURE ---->",
			   sizeof("<--- SIGNATURE ---->")-1))
	    break;
	  if (buf[0] == '\0') 
	    exit (EXIT_FAILURE);
	  sh_util_compress(bufc, buf, SH_MAXBUF-KEY_MAX);
	}
      
      /* get signature and number 
       */
      (void) sh_unix_getline (fd, key,    sizeof(key));
      key[KEY_MAX] = '\0';
      (void) sh_unix_getline (fd, number, sizeof(number));
      number[63]   = '\0';
      numsig = atol (number);
      
      fprintf(stderr, "Message %s - MAC %s\n", number, key);

      if (numsig == 0)
	{
	  sh_util_encode(key, bufc, 1);
	  strlcpy(key3, key, KEY_MAX+1);
	  fprintf (stderr, "(unchecked)\n");
	}
      else
	{
	  /* iterate key
	   */
	  strlcpy(key2, key3, KEY_MAX+1); 
	  for (i = 0; i < numsig; ++i) 
	    {
	      strlcpy (key2, sh_tiger_hash (key2, TIGER_DATA), KEY_MAX+1);
	    }
	  
	  strlcat (bufc, key2, SH_MAXBUF+1);
	  
	  if (strncmp (key, 
		       sh_tiger_hash (bufc, TIGER_DATA),
		       50) != 0) {
	    fprintf (stderr, "(FAILED)\n");
	  } else { 
	    fprintf (stderr, "(passed)\n");
	  }
	}

    } /* end scan mailbox */

  sh_unix_fileclose (fd);
  SH_FREE(buf);
  SH_FREE(bufc);

  fflush(stdout);

  _exit (EXIT_SUCCESS);
}

int sh_mail_setaddress (char * address)
{
  unsigned long  numcop;

  if (address != NULL) 
    {
      numcop = strlcpy (sh.address, address, 256+1);
      if (numcop >= 257) 
	{
	  fprintf(stderr, 
		  _("ERROR : Mail address too long (max=256): %s"), 
		  address);
	  return BAD;
	}
      return GOOD;
    }
  return BAD;
}

int sh_mail_setNum (char * str)
{
  int i = atoi (str);

  if (i >= 0 && i < 100) {
    sh.mailNum.alarm_interval = i;
    return GOOD;
  } else {
    sh_error_handle (SH_ERR_ERR, PLACE
		     NULL,
		     _("Invalid argument for maximum number of pending "\
		       "mails (%s)"),
		     str);
    return BAD;
  }
}

void sh_mail_pushstack (char * msg)
{
  mail_msg_t * new;

  if (msg == NULL || failedMail == BAD) return;

  new = SH_ALLOC(sizeof(mail_msg_t));

  /* write OUT */

  strlcpy (new->msg, msg, BUFSIZ+18+53);
  new->next = mailStack;
  mailStack = new;
  ++sh.mailNum.alarm_last;

  if (sh.mailNum.alarm_last >= sh.mailNum.alarm_interval)
    sh_mail_msg (sh.address, NULL);

  return;
}

/* the mailer
 */
/*  static void sh_exec_mail (char * address); */
static void sh_exec_mail2 (char * address);

/* return a pipe to mailer
 */
static int mail (char *address);

#ifdef HAVE_MLOCK
#include <sys/mman.h>
#endif

int sh_mail_msg (char * address, char * message)
{
    int          fdmail;
    char         subject[32+32+SH_MINIBUF+2+3];
    mail_msg_t * mailPtr;
    int          status;
    char       * bufcompress;
    static int   failcount = 0;
    static int   isfirst   = 1;
    static int   mailcount = 0;

    struct  sigaction  old_act;
    struct  sigaction  new_act;

    /* return if we cannot mail
     */
    if (failedMail == BAD) 
      return EXIT_FAILURE;

    new_act.sa_handler = SIG_IGN;

    if ( (fdmail = mail(address)) < 0) /* cannot mail */
      return EXIT_FAILURE;

    strlcpy(subject, "Subject: ",               sizeof(subject));
    strlcat(subject, sh_unix_time (0),          sizeof(subject));
    strlcat(subject, " ",                       sizeof(subject));
    strlcat(subject, sh.localhost,              sizeof(subject));
    strlcat(subject, "\n\n",                    sizeof(subject));

    /* catch (ignore) 'broken pipe'
     */
    sigaction (SIGPIPE, &new_act, &old_act);

    /* start writing to pipe
     */
    status = sh_unix_filewrite (fdmail, subject, strlen(subject), 
				SH_ISDATA, SH_DATA_RAW);

    /* ---------- Message ---------------------------------------- */

    strlcpy(subject, sh_unix_time (0),          sizeof(subject));
    strlcat(subject, " ",                       sizeof(subject));
    strlcat(subject, sh.localhost,              sizeof(subject));
    strlcat(subject, "\n",                      sizeof(subject));

    bufcompress = SH_ALLOC (SH_MAXBUF);
    bufcompress[0] = '\0';
#ifdef HAVE_MLOCK
    if (skey->mlock_failed == GOOD) 
      mlock(bufcompress , SH_MAXBUF);
#endif

    if (status == EXIT_SUCCESS)
      status = sh_unix_filewrite (fdmail, "<--- MESSAGE   ---->\n", 
				  sizeof ("<--- MESSAGE   ---->\n")-1, 
					  SH_ISDATA, SH_DATA_RAW);

    if (status == EXIT_SUCCESS)
      status = sh_unix_filewrite (fdmail, subject, strlen(subject),
				  SH_ISDATA, SH_DATA_RAW);
    sh_util_compress (bufcompress, subject, SH_MAXBUF-KEY_MAX-1);

    if (message != NULL)
      {
	if (status == EXIT_SUCCESS) 
	  {
	    status = sh_unix_filewrite (fdmail, message, strlen(message), 
				  SH_ISDATA, SH_DATA_LINE);
	    sh_util_compress (bufcompress, message, SH_MAXBUF-KEY_MAX-1);
	  } else {
	    sh_mail_pushstack (message);  /* save for later */
	    ++failcount;
	    if (failcount > SH_MAX_FAIL) 
	      failedMail = BAD;
	  }
      }

    while (mailStack != NULL && status == EXIT_SUCCESS) 
      {
	status = sh_unix_filewrite (fdmail, 
				    mailStack->msg, strlen(mailStack->msg), 
				    SH_ISDATA, SH_DATA_LINE);
	sh_util_compress (bufcompress, mailStack->msg, SH_MAXBUF-KEY_MAX-1);
	mailPtr   = mailStack;
	mailStack = mailStack->next;
	SH_FREE(mailPtr);
	--sh.mailNum.alarm_last;
      }

    /* ------ signature block ---------------------------------- */

    if (status == EXIT_SUCCESS)
      status = sh_unix_filewrite (fdmail, "<--- SIGNATURE ---->\n", 
				  sizeof ("<--- SIGNATURE ---->\n")-1, 
				  SH_ISDATA, SH_DATA_RAW);

    /* generate new signature key
     */
    if (isfirst == 1) {
      sh_util_keyinit (skey->mailkey_old, KEY_MAX+1);
    }
    strlcpy(skey->mailkey_new, 
	    sh_tiger_hash (skey->mailkey_old, TIGER_DATA), 
	    KEY_MAX+1);

    /* First message is not signed. Next message is signed with the hash
     * of the signature key.
     */
    if (isfirst == 0) 
      {
	/* sign the message with the signature key
	 */
	strlcat (bufcompress, skey->mailkey_new, SH_MAXBUF);
	
	if (status == EXIT_SUCCESS)
	  status = sh_unix_filewrite (fdmail, 
				      sh_tiger_hash (bufcompress, TIGER_DATA), 
				      KEY_MAX, 
				      SH_ISDATA, SH_DATA_RAW);
      } 
    else 
      {
	/* reveal first signature key
	 */
	strlcpy(skey->crypt, skey->mailkey_new, KEY_MAX+1);
	sh_util_encode(skey->crypt, bufcompress, 0);
	if (status == EXIT_SUCCESS)
	  status = sh_unix_filewrite (fdmail, skey->crypt, 
				      KEY_MAX, SH_ISDATA, SH_DATA_RAW);
	memset (skey->crypt, 'X', KEY_MAX);
	isfirst = 0;
      }

    /* X(n) -> X(n-1)
     */
    strlcpy (skey->mailkey_old, skey->mailkey_new, KEY_MAX+1); 

    if (status == EXIT_SUCCESS)
      status = sh_unix_filewrite (fdmail, "\n", 1, 
				  SH_ISDATA, SH_DATA_RAW);

    sprintf(subject, "%06d\n", mailcount);             /* known to fit  */
    if (status == EXIT_SUCCESS)
      status = sh_unix_filewrite (fdmail, subject, strlen(subject), 
				  SH_ISDATA, SH_DATA_RAW);
    ++mailcount;

    if (status == EXIT_SUCCESS)
      status = sh_unix_filewrite (fdmail, "<--- END       ---->\n", 
				  sizeof ("<--- END       ---->\n")-1, 
					  SH_ISDATA, SH_DATA_RAW);

    /* ------------- End -------------------------------------- */

#ifdef HAVE_MLOCK
    memset(bufcompress, 'X', SH_MAXBUF);
    if (skey->mlock_failed == GOOD) 
      munlock(bufcompress , SH_MAXBUF);
    SH_FREE(bufcompress);
#endif

    close (fdmail);

    /* reset signal
     */
    sigaction (SIGPIPE, &old_act, NULL);

    if (status != EXIT_SUCCESS) {
      sh_error_handle (SH_ERR_FATAL, PLACE
		       NULL,
		       _("Mail error - broken pipe."));
      return EXIT_FAILURE;
    } 

    return EXIT_SUCCESS;
}


static int mail (char *address)
{
  int error;
  int filedes[2];
  int pid;
  int fd;
  int fdlimit = sysconf(_SC_OPEN_MAX);

#ifdef SIGCHLD
  struct sigaction act;
#endif
 
  struct passwd * tempres;

  if (pipe(filedes) < 0) {
    failedMail = BAD;
    error = errno;
    sh_error_handle (SH_ERR_FATAL, PLACE
		     _("Mail error"),
		     _("pipe() failed: %s"),
		     sh_error_message (error));
    return (-1);
  }


  if ((pid = fork()) < 0) {
    failedMail = BAD;
    error = errno;
    close(filedes[0]);
    close(filedes[1]);
    sh_error_handle (SH_ERR_FATAL, PLACE
		     _("Mail error"),
		     _("fork() failed: %s"),
		     sh_error_message (error));
    return (-1);
  }


  /* --------  CHILD ---------------------- */

  if (pid == 0) { 

    memset(&skey, 'X', sizeof(sh_key_t));

    /* close all open files but the read pipe */
    fd = 0;
    while (fd < fdlimit) {
      if (fd != filedes[0]) close(fd);
      ++fd;
    }

    /* this should set fd 0 = in, fd 1,2 = /dev/null */
    if (STDIN_FILENO != filedes[0]) {
      dup2 (filedes[0], STDIN_FILENO); 
      close(filedes[0]);
    }
    open("/dev/null",O_RDWR);
    dup(0); dup(0);

    if (0 == geteuid()) {  
      /* we are root, so lets drop 'root' 
       */
      tempres = getpwnam(DEFAULT_IDENT);
      if (NULL != tempres) {
	setgid(tempres->pw_gid);  setuid(tempres->pw_uid);
      } else {
	perror ("sh_mail - drop root");
	exit (EXIT_FAILURE);
      }
    }



    sh_exec_mail2 (address);

    _exit (EXIT_FAILURE); /* should not be reached */
  }

  /* --------  PARENT ---------------------- */

  /* enable automatic reaping of child process 
   */
#ifdef SIGCHLD
  act.sa_handler = SIG_IGN;
#ifdef SA_NOCLDSTOP
  act.sa_flags = SA_NOCLDSTOP;
#else
  act.sa_flags = 0;
#endif
  sigemptyset(&act.sa_mask);
  sigaction(SIGCHLD, &act, NULL);
#endif

  close(filedes[0]);

  /* return the write pipe (= stdin of mail process) 
   */
  return (filedes[1]);
}


/*
 *
 * SMTP CODE BELOW
 *
 *
 */

#include <ctype.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

static int sh_mail_wait(int code, FILE * m_socket);


static void sh_exec_mail2 (char * address)
{

  char         this_address[256];
  char         m_machine[256];
  char         m_user[256];
  register int i, j, k;

  FILE               * connFile;
  int                  fd;
  struct in_addr       haddr;
  struct sockaddr_in   saddr;
  struct hostent     * hent = NULL; /* make compiler happy */
  struct hostent     * this_host;
  int                  c;
  char               * testit;
  int                  testflag = 0;

#ifdef MAIL_DEBUG
  FILE               * ErrFile;
  ErrFile = 
    fopen(DEFAULT_MAILDEBUGFILE,
	  "a");
#endif

  /* -------   split adress ------------------  */
  if (strchr (address, '@') == NULL) {
    strlcpy(m_user,    address,     256);
    strlcpy(m_machine, "localhost", 256);
  } else {
    i = 0;
    while (i < 255 && address[i] != '@') {
      m_user[i] = address[i];
      ++i;
    }
    /* adress[i] = '@' */
    m_user[i] = '\0';
    j = i + 1; k = i; i = 0;
    while (i < 255 && address[i+j] != '\0') {
      m_machine[i] = address[i+j];
      ++i;
    }

    m_machine[i] = '\0';
    if (address[k] != '@' || address[k+i+1] != '\0') {
      _exit (EXIT_FAILURE);
    } 
  }

#ifdef MAIL_DEBUG
  fprintf(ErrFile,  "user %s machine %s\n", m_user, m_machine); 
  fflush(ErrFile);
#endif

  /* check whether it is in dotted number format
   */
  testit = &m_machine[0]; testflag = GOOD;
  while (*testit != '\0') 
    {
      if (!isdigit(*testit) && *testit != '.') testflag = BAD;
      ++testit;
    }

#ifdef MAIL_DEBUG
  if (testflag == BAD)
    fprintf(ErrFile,  "no number-dot address: %s\n", m_machine);
  else
    fprintf(ErrFile,  "is a number-dot address: %s\n", m_machine);
  fflush(ErrFile);
#endif


  /* host entry of other machine
   */
  if (testflag == GOOD)
    {
      if ( 0 == inet_aton(m_machine, &haddr) ) 
	{
#ifdef MAIL_DEBUG
	  fprintf(ErrFile,  
		  "inet_aton(%s) failed\n", m_machine); fflush(ErrFile);
#endif
	  return;
	}
    }
  else
    {
      if ( (hent = gethostbyname (m_machine)) == NULL) {
#ifdef MAIL_DEBUG
	fprintf(ErrFile,  "gethostbyname(%s) failed\n", m_machine); 
	fflush(ErrFile);
#endif
	return;
      }
    }

  (void) memset (&saddr, 0, sizeof(saddr));
  if (testflag == GOOD)
    saddr.sin_addr   = haddr;
  else
    saddr.sin_addr   = *(struct in_addr *) hent->h_addr;
  saddr.sin_family = AF_INET;
  saddr.sin_port   = htons (IPPORT_SMTP);

  /* create a socket for communication
   */
  if ((fd = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1) {
#ifdef MAIL_DEBUG
  fprintf(ErrFile,  "socket() failed\n"); fflush(ErrFile);
#endif
    return;
  }


  /* associate a FILE structure with it
   */
  if ((connFile = fdopen (fd, "r+")) == NULL) {
#ifdef MAIL_DEBUG
  fprintf(ErrFile,  "fdopen() failed\n"); fflush(ErrFile);
#endif
    close(fd);
    return;
  }

  /* connect our socket to another socket somewhere else
   */
  if (connect (fd, (struct sockaddr *) &saddr, sizeof(saddr))) {
#ifdef MAIL_DEBUG
  fprintf(ErrFile,  "connect() failed\n"); fflush(ErrFile);
#endif
    fclose(connFile);
    return;
  }



  /* say HELO to the other socket
   */
  if (!sh_mail_wait (220, connFile)) {
    fclose(connFile);
    return;
  }
#ifdef MAIL_DEBUG
  fprintf(ErrFile,  "HELO %s \r\n", m_machine); fflush(ErrFile);
#endif
  fprintf(connFile, "HELO %s \r\n", m_machine);
  fflush(connFile);
  if (!sh_mail_wait(250, connFile)) {
#ifdef MAIL_DEBUG
  fprintf(ErrFile,  "timeout\n"); fflush(ErrFile);
#endif
    fclose(connFile);
    return;
  }


  /* tell them who we are
   */
  strlcpy (this_address, DEFAULT_SENDER, 256);
  strlcat (this_address, "@", 256);
  this_host = gethostbyaddr((const char *) &(saddr.sin_addr),
			    sizeof(saddr.sin_addr), AF_INET);
  if (this_host) {
    strlcat (this_address, this_host->h_name, 256);
  } else {
    /* we failed to look up our own name. */
#ifdef HAVE_VSNPRINTF
    snprintf(this_address, 255, "%s@%s", 
	     DEFAULT_SENDER, inet_ntoa(saddr.sin_addr));
#else
    if (sizeof(DEFAULT_SENDER) + strlen(inet_ntoa(saddr.sin_addr)) < 255)
      sprintf(this_address, "%s@%s",                   /* known to fit  */
	      DEFAULT_SENDER, inet_ntoa(saddr.sin_addr));
    else
      sprintf(this_address, "unknown@unknown");        /* known to fit  */
#endif
  } 

#ifdef MAIL_DEBUG
  fprintf(ErrFile,  "MAIL FROM:<%s>\r\n", this_address); fflush(ErrFile);
#endif
  fprintf(connFile, "MAIL FROM:<%s>\r\n", this_address);
  fflush(connFile);
  if (!sh_mail_wait(250, connFile)) {
#ifdef MAIL_DEBUG
  fprintf(ErrFile,  "timeout\n"); fflush(ErrFile);
  fclose(ErrFile);
#endif
    fclose(connFile);
    return;
  }

  /* tell them who to send mail to
   */
#ifdef MAIL_DEBUG
  fprintf(ErrFile,  "RCPT TO:<%s>\r\n", address); fflush(ErrFile);
#endif
  fprintf(connFile, "RCPT TO:<%s>\r\n", address); 
  fflush(connFile);
  if (!sh_mail_wait(250, connFile)) {
#ifdef MAIL_DEBUG
  fprintf(ErrFile,  "timeout\n"); fflush(ErrFile);
#endif
    fclose(connFile);
    return;
  }

  /* Send the message 
   */
#ifdef MAIL_DEBUG
  fprintf(ErrFile,  "DATA\n"); fflush(ErrFile);
#endif
  fputs("DATA\n", connFile);      
  fflush(connFile);
  if (!sh_mail_wait(354, connFile)) {
#ifdef MAIL_DEBUG
  fprintf(ErrFile,  "timeout\n"); fflush(ErrFile);
#endif
    fclose(connFile);
    return;
  }

#ifdef MAIL_DEBUG
  fprintf(ErrFile,  
	  "From: <%s>\r\n"\
	  "Date: %s\r\n",
	  this_address,
	  sh_unix_time (0)); fflush(ErrFile);
#endif
  fprintf(connFile,
	  "From: <%s>\r\n"\
	  "Date: %s\r\n",
	  this_address,
	  sh_unix_time (0));

  /* read STDIN and write to socket
   */
#ifdef MAIL_DEBUG
  fprintf(ErrFile,  "message writing starts\n"); fflush(ErrFile);
#endif
  while ( (c = getchar()) != EOF) fputc (c, connFile);
#ifdef MAIL_DEBUG
  fprintf(ErrFile,  "message written\n"); fflush(ErrFile);
#endif

  /* end connection
   */
  fputs("\r\n.\r\n", connFile);   /* End message, see RFC 821 */
  fflush(connFile);
#ifdef MAIL_DEBUG
  fprintf(ErrFile,  "message end written\n"); fflush(ErrFile);
#endif

  if (sh_mail_wait(250, connFile))
    fputs("QUIT\r\n",connFile);     /* Exit nicely */
#ifdef MAIL_DEBUG
  fprintf(ErrFile,  "exit\n"); fflush(ErrFile);
#endif
    
  _exit (EXIT_SUCCESS);
}

static jmp_buf wait_timeout;

static void sh_mail_alarmhandle (int signal)
{
  struct  sigaction  def_act;
  def_act.sa_handler = SIG_IGN;

  alarm(0);
  sigaction (SIGALRM, &def_act, NULL);
  longjmp(wait_timeout, 1);
}

static int sh_mail_wait(int code, FILE * m_socket)
{
  int rcode, g;

  enum { 
    WAIT_CODE_START, 
    WAIT_CODE, 
    WAIT_NL, 
    WAIT_NL_CONT 
  } state;

  struct  sigaction  def_act;
  struct  sigaction  new_act;
  def_act.sa_handler = SIG_IGN;
  new_act.sa_handler = sh_mail_alarmhandle;

  
  /* alarm was triggered
   */
  if (setjmp(wait_timeout) != 0)
    return 0;

  /* timeout after 5 minutes
   */
  alarm(0);
  sigaction (SIGALRM, &new_act, NULL);
  alarm(300);

  rcode = 0;
  state = WAIT_CODE_START;

  while (!feof(m_socket) && !ferror(m_socket)) {

    if ( (g=fgetc(m_socket)) == EOF)
      return 0; 

    switch(state) {

    case WAIT_CODE_START:
      if (isspace(g))
	break;             /* Skip white space */
      if (!isdigit(g))
	return 0;          /* Oops! No leading number */
      rcode = g-'0';
      state = WAIT_CODE;
      break;
      
    case WAIT_CODE:
      if (isdigit(g)) {
	rcode = rcode*10 + g - '0';
	break;
      }
      state = ((g == '-') ?  WAIT_NL_CONT :  WAIT_NL); /* See RFC 821 */
      break;
      
    case WAIT_NL:
      if (g != '\n')
	break;
      alarm(0);                   /* Disable alarm */
      sigaction (SIGALRM, &def_act, NULL); /* Reset handler */
      return ((int)(rcode/100) == (int)(code/100)) ? 1 : 0;
      
    case WAIT_NL_CONT:
      if (g == '\n')
	state = WAIT_CODE_START;  /* There is a continuation line */
      break; 
      
    default:
      return 0;
      
    }
  }
  return 0;                      /* Failed, EOF or error on socket */
}

  


