/* SAMHAIN file system integrity testing                                   */
/* Copyright (C) 1999 Rainer Wichmann                                      */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "config.h"

#include "samhain.h"
#include "sh_error.h"
#include "sh_unix.h"
#include "sh_utils.h"
#include "sh_files.h"
#include "sh_mail.h"



typedef enum {
  SH_SECTION_LOG,
  SH_SECTION_MISC,
  SH_SECTION_READONLY,
  SH_SECTION_LOGFILES,
  SH_SECTION_LOGGROW,
  SH_SECTION_NOIGNORE,
  SH_SECTION_ALLIGNORE,
  SH_SECTION_THRESHOLD
} ShSectionType;
    
static void sh_readconfig_line (char * line);

static ShSectionType read_mode;

/* read the configuration file
 */
int sh_readconf_read ()
{
  int    fd;
  char * tmp;
  char   line[80+2];

  /* open config file, exit on failure
   */
  if ( (fd = sh_unix_fileopen (DEFAULT_CONFIGFILE, 'r', 
			       SH_ISDATA, SH_ERR_FATAL)) == -1) {
    tmp = sh_util_safe_name (DEFAULT_CONFIGFILE);
    sh_error_handle (SH_ERR_FATAL, PLACE
		     NULL,
		     _("Cannot open configuration file %s"),
		     tmp);
    SH_FREE(tmp);
    exit (EXIT_FAILURE);
  }

  while ( sh_unix_getline (fd, line, 80) > 0) {

    /* skip header etc. */
    if (line[0] == '#'  ||  
	line[0] == ' '  || 
	line[0] == '\n' || 
	line[0] == '\0' || 
	line[0] == ';'  || 
	line[0] == '/' ) continue; 
  
    /* clip off trailing white space                 */
    tmp = line + strlen( line ); --tmp;
    while( isspace( *tmp ) && tmp >= line ) *tmp-- = '\0';

    if (line[0] == '[') { 

      /* -------  starts a section  ------------  */

      if      (strncmp (line,  "[ReadOnly]", 
			sizeof("[ReadOnly]")-1) == 0) 
	read_mode = SH_SECTION_READONLY;
      else if (strncmp (line,  "[LogFiles]", 
			sizeof("[LogFiles]")-1) == 0) 
	read_mode = SH_SECTION_LOGFILES;
      else if (strncmp (line,  "[GrowingLogFiles]", 
			sizeof("[GrowingLogFiles]")-1) == 0) 
	read_mode = SH_SECTION_LOGGROW;
      else if (strncmp (line,  "[IgnoreAll]", 
			sizeof("[IgnoreAll]")-1) == 0) 
	read_mode = SH_SECTION_ALLIGNORE;
      else if (strncmp (line,  "[IgnoreNone]", 
			sizeof("[IgnoreNone]")-1) == 0) 
	read_mode = SH_SECTION_NOIGNORE;
      else if (strncmp (line,  "[Log]", 
			sizeof("[Log]")-1) == 0) 
	read_mode = SH_SECTION_LOG;
      else if (strncmp (line,  "[Misc]", 
			sizeof("[Misc]")-1) == 0) 
	read_mode = SH_SECTION_MISC;
      else if (strncmp (line,  "[EventSeverity]", 
			sizeof("[EventSeverity]")-1) == 0) 
	read_mode = SH_SECTION_THRESHOLD;
      else
	sh_error_handle (SH_ERR_ERR, PLACE
			 NULL,
			 _("Invalid section in configuration file (%s)"),
			 line);
    } else { 
      /* ------  no new section ---------- */
      sh_readconfig_line (line);
    }
  }

  sh_unix_fileclose (fd);
  return (EXIT_SUCCESS);
}

typedef enum {
  SET_MAILTIME,
  SET_FILETIME 
} ShTimerItem;

    
int sh_readconf_setTime (char * str, ShTimerItem what)
{
  unsigned long i = atoi (str);

  if (i >= 0 && i < LONG_MAX) {
    if      (what == SET_MAILTIME) sh.mailTime.alarm_interval = i;
    else if (what == SET_FILETIME) sh.fileCheck.alarm_interval  = i;
    return GOOD;
  } else {
    sh_error_handle (SH_ERR_ERR, PLACE
		     NULL,
		     _("Invalid argument for maximum time interval (%s)"),
		     str);
    return BAD;
  }
}

static void sh_readconfig_line (char * line)
{    
  char * c;
  char * tmp;
  int    i;
  int    key = 0;

  static char  *ident[] = {
    "SeverityReadOnly",
    "SeverityLogFiles",
    "SeverityGrowingLogs",
    "SeverityIgnoreNone",
    "SeverityIgnoreAll",
    "SeverityDirs",
    "SeverityFiles",
    "SeverityNames",
    NULL
  };

  static int      identnum[] = { 
  SH_ERR_T_RO,    
  SH_ERR_T_LOGS,  
  SH_ERR_T_GLOG,  
  SH_ERR_T_NOIG,  
  SH_ERR_T_ALLIG, 
  SH_ERR_T_DIR,   
  SH_ERR_T_FILE, 
  SH_ERR_T_NAME        
  };
    

  /* interpret line */
  c = strchr(line, '=');
  if (c == NULL) return; /* no 'key=value' pair  */
  ++c;
  if ((*c) == '\0' || (*c) == ' ') return; /* no value */

  /* clip off trailing white space                 */
  tmp = line + strlen( line ); --tmp;
  while( isspace( *tmp ) && tmp >= line ) *tmp-- = '\0';

  sh_error_handle (SH_ERR_ALL, PLACE
		   NULL, _("ConfigFile: %s"), line);

  /* file or directory                             */
  if      (strncmp (line, "dir",  sizeof("dir")-1)  == 0) 
    key = 2;
  else if (strncmp (line, "file", sizeof("file")-1) == 0) 
    key = 1;
  else
    key = 0;

    
  if      (read_mode == SH_SECTION_READONLY) {
    if      (key == 1) sh_files_pushfile_ro (c);
    else if (key == 2) sh_files_pushdir_ro  (c);
  }
  else if (read_mode == SH_SECTION_LOGFILES) {
    if      (key == 1) sh_files_pushfile_log (c);
    else if (key == 2) sh_files_pushdir_log  (c);
  }
  else if (read_mode == SH_SECTION_LOGGROW) {
    if      (key == 1) sh_files_pushfile_glog (c);
    else if (key == 2) sh_files_pushdir_glog  (c);
  }
  else if (read_mode == SH_SECTION_NOIGNORE) {
    if      (key == 1) sh_files_pushfile_noig (c);
    else if (key == 2) sh_files_pushdir_noig  (c);
  }
  else if (read_mode == SH_SECTION_ALLIGNORE) {
    if      (key == 1) sh_files_pushfile_allig (c);
    else if (key == 2) sh_files_pushdir_allig  (c);
  }

  else if (read_mode == SH_SECTION_MISC) {
    if (strncmp (line,  "SetFilecheckTime", 
		 sizeof("SetFilecheckTime")-1) == 0) {
      sh_readconf_setTime (c, SET_FILETIME);
    }
    else if (strncmp (line,  "SetMailTime", 
		      sizeof("SetMailTime")-1) == 0) {
      sh_readconf_setTime (c, SET_MAILTIME);
    }
    else if (strncmp (line,  "SetMailNum", 
		      sizeof("SetMailNum")-1) == 0) {
      sh_mail_setNum (c);
    }
    else if (strncmp (line,  "SetMailAddress", 
		      sizeof("SetMailAddress")-1) == 0) {
      sh_mail_setaddress (c);
    }
    else if (strncmp (line,  "SetTimeServer", 
		      sizeof("SetTimeServer")-1) == 0) {
      sh_unix_settimeserver (c);
    }
    else if (strncmp (line,  "SetLoopTime", 
		      sizeof("SetLoopTime")-1) == 0) {
      sh_util_setlooptime (c);
    }
    else if (strncmp (line,  "SetRecursionLevel", 
		      sizeof("SetRecursionLevel")-1) == 0) {
      sh_files_setrecursion (c);
    }
    else if (strncmp (line,  "ChecksumTest", 
		      sizeof("ChecksumTest")-1) == 0) {
      sh_util_setchecksum (c);
    }
    
  }
  else if (read_mode == SH_SECTION_LOG) {
    if      (strncmp (line,  "DebugLevel", 
		      sizeof("DebugLevel")-1) == 0) 
      sh_error_setdebug (c);
    else if (strncmp (line,  "MailSeverity", 
		      sizeof("MailSeverity")-1) == 0)
      sh_error_setseverity (c);
    else if (strncmp (line,  "PrintSeverity", 
		      sizeof("PrintSeverity")-1) == 0)
      sh_error_setprint (c);
    else if (strncmp (line,  "LogSeverity", 
		      sizeof("LogSeverity")-1) == 0)
      sh_error_setlog (c);
    else if (strncmp (line,  "SyslogSeverity", 
		      sizeof("SyslogSeverity")-1) == 0)
      sh_error_set_syslog (c);
  }
  
  else if (read_mode == SH_SECTION_THRESHOLD) {
    i = 0;
    while (ident[i] != NULL) {
      if (strncmp (line, ident[i], strlen(ident[i])-1) == 0)
	sh_error_set_iv (identnum[i], c);
      ++i;
    }
  }
     
  return;
}
  
    
