/* SAMHAIN file system integrity testing                                   */
/* Copyright (C) 1999 Rainer Wichmann                                      */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"

#ifdef TM_IN_SYS_TIME
#include <sys/time.h>
#else
#include <time.h>
#endif

#ifdef  HAVE_UNISTD_H
#include <errno.h>
#include <signal.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/resource.h>
#ifdef HAVE_SYS_VFS_H
#include <sys/vfs.h>
#endif
#include <fcntl.h>
#include <unistd.h>
#endif

#ifdef HAVE_MLOCK
#include <sys/mman.h>
#endif


#include "samhain.h"
#include "sh_error.h"
#include "sh_unix.h"
#include "sh_utils.h"
#include "sh_mem.h"

#ifndef HAVE_LSTAT
#define lstat((x),(y)) stat(x,y)
#endif 


extern char **environ;


void sh_unix_sigexit (int signal)
{
  static char msg[62];

  strlcpy (msg, "Exit samhain on signal : ", 26+1);
#ifdef HAVE_STRSIGNAL
  strlcat (msg, strsignal(signal), sizeof(msg));
#endif

  /* cleanup is done by atexit(); do not reinstall signal handler
   * to avoid race condition
   */
  sh_error_handle (SH_ERR_STAMP, PLACE 
		   NULL,
		   msg);

#ifdef SIGSEGV
  if (signal == SIGSEGV) _exit(EXIT_FAILURE);
#endif
#ifdef SIGILL
  if (signal == SIGILL)  _exit(EXIT_FAILURE);
#endif
#ifdef SIGBUS
  if (signal == SIGBUS)  _exit(EXIT_FAILURE);
#endif
#ifdef SIGFPE
  if (signal == SIGFPE)  _exit(EXIT_FAILURE);
#endif

  exit(EXIT_FAILURE);
}

void sh_unix_siginstall ()
{
  struct sigaction act, oldact; 

  act.sa_handler = &sh_unix_sigexit;    /* signal action           */
  sigemptyset( &act.sa_mask );          /* set an empty mask       */
#ifdef SA_RESETHAND
  act.sa_flags = SA_RESETHAND;          /* init sa_flags           */
#else
  act.sa_flags = 0;                     /* init sa_flags           */
#endif

  /* signals that may cause termination */

#ifdef SIGHUP
  sigaction(SIGHUP,  &act, &oldact);
#endif
#ifdef SIGINT
  sigaction(SIGINT,  &act, &oldact);
#endif
#ifdef SIGQUIT
  sigaction(SIGQUIT, &act, &oldact);
#endif
#ifdef SIGILL
  sigaction(SIGILL,  &act, &oldact);
#endif
#ifdef SIGABRT
  sigaction(SIGABRT, &act, &oldact);
#endif
#ifdef SIGSEGV
  sigaction(SIGSEGV, &act, &oldact);
#endif
#ifdef SIGTRAP
  sigaction(SIGTRAP, &act, &oldact);
#endif
#ifdef SIGIOT
  sigaction(SIGIOT,  &act, &oldact);
#endif
#ifdef SIGTERM
  sigaction(SIGTERM, &act, &oldact);
#endif
#ifdef SIGBUS
  sigaction(SIGBUS,  &act, &oldact);
#endif
#ifdef SIGIO
  sigaction(SIGIO,   &act, &oldact);
#endif
#ifdef SIGPOLL
  sigaction(SIGPOLL, &act, &oldact);
#endif
#ifdef SIGXCPU
  sigaction(SIGXCPU, &act, &oldact);
#endif
#ifdef SIGPWR
  sigaction(SIGPWR,  &act, &oldact);
#endif
#ifdef SIGFPE
  sigaction(SIGFPE,  &act, &oldact);
#endif


  return;
}
  

/* drop SUID privileges
 */
int sh_unix_droppriv()
{
  setgid(getgid());
  return setuid(getuid());
}

/* determine file type
 */
static int sh_unix_getinfo_type (struct stat * buf, 
				 ShFileType * type, char * c_mode)
{
  if      ( S_ISREG(buf->st_mode)  ) { 
    (*type)   = SH_FILE_REGULAR;
    c_mode[0] = '-';
  }
  else if ( S_ISLNK(buf->st_mode)  ) {
    (*type)   = SH_FILE_SYMLINK;
    c_mode[0] = 'l';
  }
  else if ( S_ISDIR(buf->st_mode)  ) {
    (*type)   = SH_FILE_DIRECTORY;
    c_mode[0] = 'd';
  }
  else if ( S_ISCHR(buf->st_mode)  ) {
    (*type)   = SH_FILE_CDEV;
    c_mode[0] = 'c';
  }
  else if ( S_ISBLK(buf->st_mode)  ) {
    (*type)   = SH_FILE_BDEV;
    c_mode[0] = 'b';
  }
  else if ( S_ISFIFO(buf->st_mode) ) {
    (*type)   = SH_FILE_FIFO;
    c_mode[0] = '|';
  }
  else if ( S_ISSOCK(buf->st_mode) ) {
    (*type)   = SH_FILE_SOCKET;
    c_mode[0] = 's';
  }
  else                              {
    (*type)   = SH_FILE_UNKNOWN;
    c_mode[0] = '?';
  }

  return EXIT_SUCCESS;
}


static int  sh_unix_getinfo_mode (struct stat *buf, 
				  unsigned int * mode, char * c_mode)
{
  (*mode) = buf->st_mode;

  /* make 'ls'-like string */
  
  if ( (buf->st_mode & S_IRUSR) != 0 )  c_mode[1] = 'r'; 
  if ( (buf->st_mode & S_IWUSR) != 0 )  c_mode[2] = 'w'; 
  if ( (buf->st_mode & S_IXUSR) != 0 ) {
    if ((buf->st_mode & S_ISUID) != 0 ) c_mode[3] = 's';
    else                                c_mode[3] = 'x';
  } else {
    if ((buf->st_mode & S_ISUID) != 0 ) c_mode[3] = 'S';
  }

  if ( (buf->st_mode & S_IRGRP) != 0 )  c_mode[4] = 'r'; 
  if ( (buf->st_mode & S_IWGRP) != 0 )  c_mode[5] = 'w'; 
  if ( (buf->st_mode & S_IXGRP) != 0 )  {
    if ((buf->st_mode & S_ISGID) != 0 ) c_mode[6] = 's';
    else                                c_mode[6] = 'x';
  } else {
    if ((buf->st_mode & S_ISGID) != 0 ) c_mode[6] = 'S';
  } 

  if ( (buf->st_mode & S_IROTH) != 0 )  c_mode[7] = 'r'; 
  if ( (buf->st_mode & S_IWOTH) != 0 )  c_mode[8] = 'w';
#ifdef S_ISVTX  /* not POSIX */
  if ( (buf->st_mode & S_IXOTH) != 0 )  {
    if ((buf->st_mode & S_ISVTX) != 0 ) c_mode[9] = 't';
    else                                c_mode[9] = 'x';
  } else {
    if ((buf->st_mode & S_ISVTX) != 0 ) c_mode[9] = 'T';
  }
#else
  if ( (buf->st_mode & S_IXOTH) != 0 )  c_mode[9] = 'x';
#endif 

  return EXIT_SUCCESS;
}


static void sh_unix_closeall (int fd)
{
    int fdlimit = sysconf(_SC_OPEN_MAX);

    while (fd < fdlimit)
      close(fd++);
    return;
}

static void sh_unix_setlimits()
{
  struct rlimit limits;

  limits.rlim_cur = RLIM_INFINITY;
  limits.rlim_max = RLIM_INFINITY;

  setrlimit (RLIMIT_CPU,     &limits);
  setrlimit (RLIMIT_FSIZE,   &limits);
  setrlimit (RLIMIT_DATA,    &limits);
  setrlimit (RLIMIT_STACK,   &limits);
  setrlimit (RLIMIT_RSS,     &limits);
  setrlimit (RLIMIT_NPROC,   &limits);
  setrlimit (RLIMIT_NOFILE,  &limits);
  setrlimit (RLIMIT_MEMLOCK, &limits);

  /* no core dumps
   */
  limits.rlim_cur = 0;
  limits.rlim_max = 0;
  setrlimit (RLIMIT_CORE,    &limits);

  return;
}

/* delete all environment variables
 */
static void sh_unix_zeroenv()
{
  char * c;
  char ** env = environ;

  while (*env != NULL) {
    c = strchr ((*env), '=');
    ++c;
    ++(env);
    while ((*c) != '\0') {
      (*c) = '\0';
      ++c;
    }
  }
  return;
}


static void  sh_unix_resettimer()
{
  struct itimerval this_timer;

  this_timer.it_value.tv_sec  = 0;
  this_timer.it_value.tv_usec = 0;

  setitimer(ITIMER_REAL,    &this_timer, NULL);
  setitimer(ITIMER_VIRTUAL, &this_timer, NULL);
  setitimer(ITIMER_PROF,    &this_timer, NULL);

  return;
}

static void  sh_unix_resetsignals()
{
  int  sig_num;
#ifdef NSIG
  int  max_sig = NSIG;  /* FIXME is NSIG the maximum number ? */
#else
  int  max_sig = 255;
#endif
  int  test;
  struct sigaction act, oldact;
  int  status;
  
  act.sa_handler = SIG_DFL;         /* signal action           */
  sigemptyset( &act.sa_mask );      /* set an empty mask       */
  act.sa_flags = 0;                 /* init sa_flags           */

  for (sig_num = 1; sig_num <= max_sig; ++sig_num) 
    {
      test = sigaction(sig_num,  &act, &oldact);
      if ((test == -1) && (errno != EINVAL)) 
	{
	  status = errno;
	  sh_error_handle (SH_ERR_WARN, PLACE
			   _("Initialization failed"),
			   _("Failed reset of signal %d: %s"),
			   sig_num,
			   sh_error_message (status));
	}
    }

  return;
}

void sh_unix_localhost()
{
  /* local hostname
   */
  (void) gethostname(sh.localhost, SH_MINIBUF);
}

void sh_unix_mlock()
{
#ifdef HAVE_MLOCK
  if (skey->mlock_failed == GOOD)
    {
      if ( (-1) == mlock( (char *) skey, sizeof (sh_key_t)) ) 
	{
	  skey->mlock_failed = BAD;
	}
    }
#else
  if (skey->mlock_failed = GOOD)
    {
      skey->mlock_failed = BAD;
    }
#endif
}  
/* daemon mode 
 */
int sh_unix_setdeamon(char * dummy)
{
  sh.modus = 1;
  return 0;
}

int sh_unix_init(int goDaemon)
{
  int status;

  /* fork twice, exit the parent process
   */
  if (goDaemon == 1) {
    
    switch (fork()) {
    case 0:  break;      /* child process continues */
    case -1: return -1;  /* error                   */
    default: _exit(0);   /* parent preocess exits   */
    }

    /* Child processes do not inherit page locks across a fork.
     * Error in next fork would return in this (?) thread of execution.
     */
    sh_unix_mlock();
    setsid();            /* should not fail         */

    switch (fork()) {
    case 0:  break;      /* child process continues */
    case -1: return -1;  /* error                   */
    default: _exit(0);   /* parent preocess exits   */
    }

    /* Child processes do not inherit page locks across a fork.
     */
    sh_unix_mlock();

  } else {
    setsid();            /* should not fail         */
  } 

  /* set working directory   
   */
  status = chdir("/");
  if ( (-1) == status ) 
    {
      sh_error_handle (SH_ERR_WARN, PLACE
		       _("Chdir failed"),
		       _("Failed chdir(): %s"),
		       sh_error_message (status));
    }
      

  /* reset timers 
   */
  sh_unix_resettimer();

  /* signal handlers 
   */
  sh_unix_resetsignals();
  sh_unix_siginstall ();

  /* set resource limits to maximum 
   */
  sh_unix_setlimits();

  /* set file creation mask 
   */
  (void) umask (0); /* should not fail */

  /* zero out the environment (like PATH='\0')  
   */
  sh_unix_zeroenv();

  /* close all file descriptors 
   */
  if (goDaemon == 1) sh_unix_closeall (0);

  /* open first three streams 
   */
  if (goDaemon == 1) {
    open("/dev/null",O_RDWR);
    dup(0); 
    dup(0);
  }

  return 0;
}

unsigned long sh_unix_longtime ()
{
  time_t        time_now;

  (void) time (&time_now);
  return (unsigned long)time_now;
} 

int sh_unix_settimeserver (char * address)
{
  unsigned long  numcop;

  if (address != NULL) 
    {
      numcop = strlcpy (sh.timeserver, address, 256+1);
      if (numcop >= 257) 
	{
	  fprintf(stderr, 
		  _("ERROR : Mail address too long (max=256): %s"), 
		  address);
	  return BAD;
	}
      return GOOD;
    }
  return BAD;
}


#ifdef HAVE_NTIME
#define UNIXEPOCH 2208988800UL  /* difference between Unix time and net time 
                                 * The UNIX EPOCH starts in 1970.
                                 */
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <ctype.h>
#endif

char * sh_unix_time (unsigned long thetime)
{
#ifdef HAVE_STRFTIME
  int           status;
#endif
  time_t        time_now;
  struct tm   * time_ptr;
  static char   AsciiTime[81];                       /* local time   */
  static char   deftime[] = "[00-00-0000 00:00:00]"; /* default time */


#ifdef HAVE_NTIME
  struct in_addr       haddr;   /* host address from numeric                */
                                /* host details returned by the DNS         */
  struct hostent *host_entry=NULL;   
  /* struct servent *time_server;  service file entry - for the port number */
  struct sockaddr_in sin;       /* socket to the remote host                */
  int    fd;                    /* network file descriptor                  */
  u_char net_time[4];           /* remote time in network format            */
  int    fail = 0;              /* no net time                              */
  static int failerr = 0;       /* no net time                              */
  fd_set fds;
  struct timeval tv;
  int    testflag;
  char  *testit;

#endif

  
  if (thetime == 0) {

#ifdef HAVE_NTIME
    if (strncmp("NULL", sh.timeserver, 4) == 0) {
#endif

      (void) time (&time_now);

#ifdef HAVE_NTIME
    } else {  /* have a timeserver address */

      /* ---- Get Time from Timeserver ---- */

      /* check whether it is in dotted number format
       */
      testit = &sh.timeserver[0]; testflag = GOOD;
      while (*testit != '\0') 
	{
	  if (!isdigit(*testit) && *testit != '.') testflag = BAD;
	  ++testit;
	}
      
      /* only use gethostbyname() if neccessary
       */
      if (testflag == GOOD)
	{
	  if ( 0 == inet_aton(sh.timeserver, &haddr) ) 
	    {
	      fail = 1;
	    }
	}
      else
	{ 
	  if ((host_entry = gethostbyname(sh.timeserver)) == NULL) {
	    if (sh.modus == 0)
	      fprintf(stderr, "Unknown host: %s.\n", sh.timeserver);
	    fail = 1;
	  } 
	}

      /****************************************************************
      else if ((time_server = getservbyname("time","tcp")) == NULL) {
	if (sh.modus == 0) 
	  fprintf(stderr, "Unknown service: time/tcp.\n");
	fail = 1;
	(void) time (&time_now);
      } 
      *****************************************************************/

      if (fail != 1) 
	{ 
	  fd = socket(AF_INET, SOCK_STREAM, 0);
	  if (fd < 0) {
	    fail = 1;
	    if (sh.modus == 0)
	      fprintf(stderr, "Error %s: socket(): %s\n", 
		      sh.timeserver, strerror(errno));
	  }
	}
      else
	{
	  fd = -1;
	}
      
      if (fail != 1) {

	/*  sin.sin_family = host_entry->h_addrtype; */
	sin.sin_family = AF_INET;
	sin.sin_port   = htons (IPPORT_TIMESERVER);
	/*  sin.sin_port   = time_server->s_port; */
	if (testflag == GOOD)
	  sin.sin_addr   = haddr;
	else
	  sin.sin_addr   = *(struct in_addr *) host_entry->h_addr;
	
	if (connect(fd, (struct sockaddr *) &sin, sizeof(sin)) < 0) 
	  {
	    if (sh.modus == 0)
	      fprintf(stderr, "Error %s: connect(): %s\n", 
		      sh.timeserver, strerror(errno));
	    close(fd);
	    fail = 1; 
	  }
	else 
	  {
	    FD_ZERO(&fds);
	    FD_SET(fd, &fds);
	    
	    tv.tv_sec  = 2;
	    tv.tv_usec = 0;

	    if (select(fd+1, &fds, NULL, NULL, &tv) > 0) {
	      if (read(fd, net_time, 4) != 4) {     
		if (sh.modus == 0)
		  fprintf(stderr, "Error: read(): %s\n", strerror(errno));
		fail = 1;
	      }
	    } else {
		if (sh.modus == 0)
		  fprintf(stderr, "Timeout expired\n");
		fail = 1;
	    } 
	    close(fd);
	  }

	if (fail == 0) 
	  { 
	    time_now = ntohl(* (long *) net_time) - UNIXEPOCH;
	    if (failerr == 1) {
	      failerr = 0;
	      sh_error_handle (SH_ERR_ERR, PLACE
			       NULL,
			       _("Connection to timeserver %s "\
				 "re-established."),
			       sh.timeserver);
	    } 
	  } 
      }

      if (fail == 1)
	{
	  (void) time (&time_now);
	  if (failerr == 0)
	    {
	      failerr = 1;
	      sh_error_handle (SH_ERR_ERR, PLACE
			       NULL,
			       _("No connection to timeserver %s."),
			       sh.timeserver);
	    }
	}
	
    }
#endif
 
  } else {
    time_now = thetime;
  }

  if (time_now == (-1) ) { 
    if (sh.modus == 0)
      fprintf(stderr, "Cannot determine time.\n");
    return &deftime[0];
  }

  time_ptr   = localtime (&time_now);

  if (time_ptr != NULL) {
#ifdef HAVE_STRFTIME
    status = strftime (AsciiTime, 80, "[%d-%m-%Y %H:%M:%S]", time_ptr);
    if ( (status == 0) || (status == 80) ) {
      if (sh.modus == 0)
	fprintf(stderr, "Cannot determine time.\n");
      return &deftime[0];
    }
    return &AsciiTime[0];
#else
    printf (AsciiTime, "[%2d-%2d-%4d %2d:%2d:%2d]", 
	    time_ptr->tm_mday, 1+time_ptr->tm_mon, 1900+time_ptr->tm_year,
	    time_ptr->tm_hour, time_ptr->tm_min, time_ptr->tm_sec);
    return &AsciiTime[0];
#endif
  } else {
    if (sh.modus == 0)
      fprintf(stderr, "Cannot determine time.\n");
    return &deftime[0];
  }

  return &deftime[0]; /* notreached */
}

/**************************************
char * sh_unix_gethome (char * name)
{
  struct passwd * tempres;

  tempres = getpwnam(name);
  if (tempres == NULL) {
    sh_error_handle (SH_ERR_ERR, PLACE
		     _("No such user: %s"),
		     _("Failed getpwnam() on %s"),
		     name);
    return NULL;
  }
  return tempres->pw_dir;
}

uid_t sh_unix_getuid (char * name)
{
  struct passwd * tempres;
  uid_t  theuid;

  tempres = getpwnam(name);
  if (tempres == NULL) {
    sh_error_handle (SH_ERR_ERR, PLACE
		     _("No such user: %s"),
		     _("Failed getpwnam() on %s"),
		     name);
    return -1;
  }

  theuid = tempres->pw_uid;
  return theuid;
}
****************************************/

int sh_unix_getUser ()
{
  struct passwd * tempres;

  tempres = getpwuid(geteuid());

  if (tempres == NULL) return BAD;

  if (tempres->pw_name != NULL) {
    if (strlen(tempres->pw_name) > SH_MINIBUF) {
      sh_error_handle (SH_ERR_ERR, PLACE
		       _("Username too long"), 
		       _("Username (%s) exceeds limit"), 
		       tempres->pw_name );
      return BAD;
    } else {
      strlcpy ( sh.user, tempres->pw_name, SH_MINIBUF+1);
    }
  } else {
    sh_error_handle (SH_ERR_ERR, PLACE
		     _("Cannot find out who you are"), 
		     _("Password field pw_user is NULL") );
    sh.user[0] = '\0';
    return BAD;
  }

  if (tempres->pw_gecos != NULL) {
    if (strlen(tempres->pw_gecos) > SH_MINIBUF) {
      sh_error_handle (SH_ERR_ERR, PLACE
		       _("Real username too long"), 
		       _("Gecos field (%s) exceeds limit"), 
		       tempres->pw_gecos );
      return BAD;
    } else {
      strlcpy ( sh.real_user, tempres->pw_gecos, SH_MINIBUF+1);
    }
  } else {
    sh_error_handle (SH_ERR_WARN, PLACE
		     _("Cannot determine real username"), 
		     _("Password field pw_gecos is NULL") );
    sh.real_user[0] = '\0';
  }


  if (tempres->pw_dir != NULL) {
    if (strlen(tempres->pw_dir) > SH_BUFSIZE) {
      sh_error_handle (SH_ERR_ERR, PLACE
		       _("Home directory too long"), 
		       _("Home directory path too long (%s)"), 
		       tempres->pw_dir );
      return BAD;
    } else {
      strlcpy ( sh.home, tempres->pw_dir, SH_BUFSIZE+1);
      return GOOD;
    }
  } else {
    sh_error_handle (SH_ERR_ERR, PLACE
		     _("Cannot determine home directory"), 
		     _("Password field pw_dir is NULL") );
    sh.home[0] = '\0';
    return BAD;
  }

  return BAD;  /* notreached */
}


char *  sh_unix_getUIDname (ShErrLevel level, uid_t uid)
{
  struct passwd * tempres;

  tempres = getpwuid(uid);

  if (tempres == NULL) {
    sh_error_handle (level, PLACE
		     _("Cannot find user for UID %ld"), 
		     _("Password file entry is NULL for UID %ld"),
		     (long) uid);
    return NULL;
  }

  if (tempres->pw_name != NULL) {
      return tempres->pw_name;
  } else {
    sh_error_handle (level, PLACE
		     _("Cannot find user for UID %ld"), 
		     _("Password field pw_user is NULL for UID %ld"),
		     (long) uid);
    return NULL;
  }
}

char *  sh_unix_getGIDname (ShErrLevel level, gid_t gid)
{
  struct group * tempres;

  tempres = getgrgid(gid);

  if (tempres == NULL) {
    sh_error_handle (level, PLACE
		     _("Cannot find group for GID %ld"), 
		     _("Group file entry is NULL for GID %ld"),
		     (long) gid);
    return NULL;
  }

  if (tempres->gr_name != NULL) {
      return tempres->gr_name;
  } else {
    sh_error_handle (level, PLACE
		     _("Cannot find group for GID %ld"), 
		     _("Group field pw_user is NULL for GID %ld"),
		     (long) gid);
    return NULL;
  }
}

int  sh_unix_fileopen (char * filename, char mode, 
		       ShOpenType IsType, ShErrLevel level)
{
  struct stat   lbuf;
  struct stat   buf;
  int           lstat_return;
  int           stat_return;
  struct passwd *filepwd;
  int           fd;
  char        * tmp;
  char        * tmpdir;
  char          message[81];
  static  int   firstaccess = 1;

  (void) umask (0);

  lstat_return = lstat(filename, &lbuf);
  if (lstat_return == -1 && errno != ENOENT) {
    lstat_return = errno;
    if (IsType == SH_ISLOG)  sh_error_logoff();
    tmp = sh_util_safe_name (filename);
    sh_error_handle (level, PLACE
		     _("Cannot check file: %s"),
		     _("Cannot lstat() file: %s: %s"),
		     tmp, 
		     sh_error_message (lstat_return));
    SH_FREE(tmp);
    return (-1);
  } else if (errno == ENOENT) lstat_return = ENOENT;
  
  switch (mode) {
  case 'w':
    tmpdir = sh_util_basename(filename);
    if (BAD == sh_unix_is_secure_dir(level, tmpdir)) {
      if (IsType == SH_ISLOG)  sh_error_logoff();
      tmp = sh_util_safe_name (filename);
      sh_error_handle (SH_ERR_FATAL, PLACE
		       NULL,
		       _("Attempt to write file (%s) in insecure dir"),
		       tmp);
      SH_FREE(tmp);
      SH_FREE(tmpdir);
      exit (EXIT_FAILURE);
    }
    SH_FREE(tmpdir);
    fd = open (filename, O_WRONLY|O_CREAT, S_IWUSR|S_IRUSR|S_IRGRP);
    break;

  case 'r':
    fd = open (filename, O_RDONLY);
    break;

  default:
    return (-1);
    break;
  } 

  
  if (fd == -1) {
    stat_return = errno;
    if (IsType == SH_ISLOG)  sh_error_logoff();
    tmp = sh_util_safe_name (filename);
    sh_error_handle (level, PLACE
		     _("Cannot open file: %s"),
		     _("Cannot open file: %s: %s"),
		     tmp,
		     sh_error_message (stat_return));
    SH_FREE(tmp);
    return (-1);
  } 
    
  /* by default, append to file
   */
  if (mode == 'w') lseek (fd, 0L, SEEK_END);


  stat_return = fstat(fd, &buf);

  if ( (stat_return == -1) ) {

    /* does not exist / permission denied          */

    stat_return = errno;
    if (IsType == SH_ISLOG)  sh_error_logoff();
    tmp = sh_util_safe_name (filename);
    sh_error_handle (level, PLACE
		     _("Cannot access file: %s"),
		     _("Cannot stat() file: %s: %s"),
		     tmp, 
		     sh_error_message (stat_return));
    SH_FREE(tmp);
    return (-1);

  } else {

    /* is it what we want ? If it had not existed before, lstat()
     * will have failed with ENOENT
     */
    if (lstat_return != ENOENT && buf.st_ino != lbuf.st_ino) {
      if (IsType == SH_ISLOG)  sh_error_logoff();
      tmp = sh_util_safe_name (filename);
      sh_error_handle (level, PLACE
		       _("Do not access because possible link: %s"),
		       _("lstat() st_ino != fstat() st_ino: %s"),
		       tmp);
      SH_FREE(tmp);
      return (-1);
    }

    /* exists, check permissions and ownership */

    stat_return = 0;

    if (mode == 'w') {

      /* check on lock
       */
      if (firstaccess == 1 && IsType == SH_ISLOG) {
	if (buf.st_nlink > 1) {
	  sh_error_logoff();
	  tmp = sh_util_safe_name (filename);
	  sh_error_handle (level, PLACE
			   _("Logfile %s is locked"),
			   _("Logfile %s is locked (%d hardlinks)"),
			   tmp, 
			   buf.st_nlink);
	  SH_FREE(tmp);
	  return (-1);

	} else if (IsType == SH_ISLOG) {
	  sh_unix_lock (filename, DEFAULT_ERRLOCK);
	  firstaccess = 0;
	}
      }
    
      if ( buf.st_uid != geteuid() ) {  /* we must be owner to write */
	/* find the owner  */
	stat_return = -1;
	filepwd = getpwuid(buf.st_uid);
	strlcpy(message, _("File owned by user: "), 81);
	strlcat(message, filepwd->pw_name, 81 );
	strlcat(message, ": %s", 81 );
      }

      if ( !S_ISREG(buf.st_mode) && 
	   (IsType == SH_ISFILE || IsType == SH_ISLOG)) {   
	/* not a regular file */
	stat_return = -1;
	strlcpy(message, _("File is not a regular file: %s"), 81);
      }

      if ( (buf.st_mode & S_IWOTH)  && (IsType == SH_ISLOG) ) {
	stat_return = -1;
	strlcpy(message, _("Logfile is world writeable: %s"), 81);
      }

    } else if (mode == 'r') {

      if ( (IsType == SH_ISDIR) && (!S_ISDIR(buf.st_mode)) ) {   
	/* is it a directory ? */
	stat_return = -1;
	strlcpy(message, _("File is not a directory: %s"), 81);
      }

      if (IsType == SH_ISDATA) {

	if ( (buf.st_mode & S_IWOTH)  && (IsType == SH_ISDATA) ) {
	  stat_return = -1;
	  strlcpy(message, _("Datafile is world writeable: %s"), 81);
	}

	if ( !S_ISREG(buf.st_mode) ) {   
	  /* not a regular file */
	  stat_return = -1;
	  strlcpy(message, _("File is not a regular file: %s"), 81);
	}

      }
      
    }

    if (stat_return == -1) {
      if (IsType == SH_ISLOG)  sh_error_logoff();
      tmp = sh_util_safe_name (filename);
      sh_error_handle (level, PLACE
		       NULL,
		       message,
		       tmp);
      SH_FREE(tmp);
      /* if (IsType == SH_ISLOG)  sh_error_logrestore(); */
      sh_unix_fileclose (fd);
      return (-1);
    }


  }

#ifdef MEM_DEBUG
  ++sh.openfiles;
  sh_mem_openf(filename, fd);
#endif

  return fd;
}
  
int sh_unix_filewrite (int fd, void * msg, long bytecount, 
		       ShOpenType IsType, ShDataType IsData)
{
  int stat_return = 0;
  int err_store   = 0;
  int try_count   = 0;
  int err_sync    = 0;
  ShErrLevel    level = SH_ERR_ERR;

  if (IsType == SH_ISLOG)  level = SH_ERR_FATAL;
  if (IsType == SH_ISDATA) level = SH_ERR_FATAL;
  if (IsType == SH_ISFILE) level = SH_ERR_ERR;
  if (IsType == SH_ISDIR)  level = SH_ERR_ERR;

  /* write message
   */
  do {

    stat_return = write (fd, msg, bytecount);
    if (stat_return == -1) { 
      err_store = errno;
      ++try_count;
    } else {
      if (IsType == SH_ISLOG) err_sync = fsync (fd);
      if (err_sync == (-1)) {
	err_sync = errno;
	sh_error_handle (level, PLACE
			 _("Could not syncronize logfile"),
			 _("fsync() error: %s"), 
			 sh_error_message (err_sync));
      }
    }

  } while ((stat_return == -1) 
	   && (err_store == EAGAIN || err_store == EINTR) 
	   && (try_count < 3));

  /* write newline if requested
   */
  if (IsData == SH_DATA_LINE) {
    do {

      stat_return = write (fd, "\n", 1);
      if (stat_return == -1) { 
	err_store = errno;
	++try_count;
      } else {
	if (IsType == SH_ISLOG) err_sync = fsync (fd);
	if (err_sync == (-1)) {
	  err_sync = errno;
	  sh_error_handle (level, PLACE
			   _("Could not syncronize logfile"),
			   _("fsync() error: %s"), 
			   sh_error_message (err_sync));
	}
      }

    } while ((stat_return == -1) 
	     && (err_store == EAGAIN  || err_store == EINTR) 
	     && (try_count < 3));
  }

  if (stat_return == -1) {
    if (IsType == SH_ISLOG)  sh_error_logoff();
    sh_error_handle (level, PLACE
		     _("A write error occurred"),
		     _("write() error: %s"), 
		     sh_error_message (err_store));
    if (IsType == SH_ISLOG)  sh_error_logrestore();
    return (EXIT_FAILURE);
  }

  return (EXIT_SUCCESS);
}

int sh_unix_getline (int fd, char * line, int sizeofline)
{
  register int  count;
  register int  n = 0;
  char          c;

  if (sizeofline < 2) {
    line[n] = '\0';
    return 0;
  }

  while (n < sizeofline) {

    count = sh_unix_fileread (fd, &c, 1);

    if (count < 1) {
      line[n] = '\0';
      break;
    } 

    if (c != '\0' && c != '\n') {
      line[n] = c;
      ++n;
    } else if (c == '\n') {
      if (n > 0) {
	line[n] = '\0';
	break;
      } else {
	line[n] = '\n'; /* get newline only if only char on line */
	++n;
	line[n] = '\0';
	break;
      }
    } else {
      line[n] = '\0';
      break;
    }

  }

  line[sizeofline] = '\0';  /* make sure line is terminated */
  return n;
}



int sh_unix_fileread (int fd, void * buf, size_t count)
{
  int status;
  int store;
  int test = 0;
  
  while (test < 3) {
    status = read (fd, buf, count);
    if (status == (-1)) {
      store = errno;
      if (errno == EINTR || errno == EAGAIN) {
	++test;
      } else {
	sh_error_handle (SH_ERR_ERR, PLACE 
			 _("Error reading file"),
			 _("Error reading file: %s"),
			 sh_error_message(store)
			 );
	return status;
      }
    } else {
      break; /* successful read */
    }
  }

  return status;
}

void sh_unix_fileclose (int fd)
{
  /* FIXME may fail on 'write-behind' systems (NFS) */
#ifdef MEM_DEBUG
  if (close (fd) != (-1)) {
    --sh.openfiles;
    return sh_mem_closef(fd);
  }
#else
  close (fd);
#endif
  return;
}


/* obtain file info
 */
int sh_unix_getinfo (ShErrLevel level, 
		     char * filename, file_type * theFile)
{
  struct stat   buf;
  struct stat   lbuf;
  int           stat_return;
  register int  i;
  ShFileType    type;
  unsigned int  mode;
  char        * name;
  char        * tmp;
  char        * tmp2;
  char        * linknamebuf;
  int           linksize;
  

  /* stat the file
   */
  stat_return = lstat (theFile->fullpath, &buf);/* FIXME lstat is not POSIX */

  /* check for error
   */
  if (stat_return != 0) {
    stat_return = errno;
    tmp2 = sh_util_safe_name (theFile->fullpath);
    sh_error_handle (level, PLACE
		     _("Could not check file %s"),
		     _("Could not check file %s: %s"),
		     tmp2,
		     sh_error_message (stat_return)
		     );
    SH_FREE(tmp2);
    return EXIT_FAILURE;
  }

  for (i = 0; i < 10; ++i) theFile->c_mode[i] = '-';
  theFile->c_mode[10] = '\0';
  for (i = 0; i < 11; ++i) theFile->link_c_mode[i] = '-';
  theFile->link_c_mode[10] = '\0';


  /* determine file type
   */
  sh_unix_getinfo_type (&buf, &type, theFile->c_mode);
  theFile->type = type;

  /* determine permissions
   */
  sh_unix_getinfo_mode (&buf, &mode, theFile->c_mode);
  theFile->mode = mode;

  /* trivia
   */
  theFile->dev       = (unsigned long) buf.st_dev;
  theFile->ino       = (unsigned long) buf.st_ino;
  theFile->mode      = (unsigned long) buf.st_mode;
  theFile->hardlinks = (unsigned long) buf.st_nlink;
  theFile->owner     = (long)          buf.st_uid;  
  theFile->group     = (long)          buf.st_gid;  
  theFile->rdev      = (unsigned long) buf.st_rdev;
  theFile->size      = (unsigned long) buf.st_size;
  theFile->blksize   = (unsigned long) buf.st_blksize;
  theFile->blocks    = (unsigned long) buf.st_blocks;
  theFile->atime     = (unsigned long) buf.st_atime;
  theFile->mtime     = (unsigned long) buf.st_mtime;
  theFile->ctime     = (unsigned long) buf.st_ctime;


  /* owner and group
   */

  if ( (name = sh_unix_getGIDname(level, buf.st_gid)) != NULL) {
    strlcpy (theFile->c_group, name, GROUP_MAX+1);
  } else {
    tmp2 = sh_util_safe_name (theFile->fullpath);
    sh_error_handle (level, PLACE
		     _("No group name for file: %s"),
		     _("No group name for file: %s"),
		     tmp2);
    SH_FREE(tmp2);
    theFile->c_group[0] = '\0';
  }

  
  if ( (name = sh_unix_getUIDname(level, buf.st_uid)) != NULL) {
    strlcpy (theFile->c_owner, name, OWNER_MAX+1);
  } else {
    tmp2 = sh_util_safe_name (theFile->fullpath);
    sh_error_handle (level, PLACE
		     _("No owner name for file: %s"),
		     _("No owner name for file: %s"),
		     tmp2);
    SH_FREE(tmp2);
    theFile->c_owner[0] = '\0';
  }

  /* output the file
   */
  tmp2 = sh_util_safe_name (filename);
  sh_error_handle (SH_ERR_ALL, PLACE
		   NULL,
		   "%10s %2d %8s %8s %14ld %21s %s", 
		   theFile->c_mode,
		   theFile->hardlinks,
		   theFile->c_owner,
		   theFile->c_group,
		   theFile->size,
		   sh_unix_time (theFile->mtime),
		   tmp2);
  SH_FREE(tmp2);

  /* check for links
   */
  if (theFile->c_mode[0] == 'l') {

    linknamebuf = SH_ALLOC(PATH_MAX + 2);

    linksize    = readlink (theFile->fullpath, linknamebuf, PATH_MAX);
    if (linksize < PATH_MAX) linknamebuf[linksize] = '\0';
    else linknamebuf[PATH_MAX] = '\0';

    if (linksize < 0) {
      linksize = errno;
      tmp2 = sh_util_safe_name (theFile->fullpath);
      sh_error_handle (level, PLACE
		       _("No link found for %s"),
		       _("Error reading link value for file %s: %s"),
		       tmp2,
		       sh_error_message (linksize));
      SH_FREE(tmp2);
      SH_FREE(linknamebuf);
      return EXIT_FAILURE;
    }
    linknamebuf[linksize] = '\0';
    
    if (linknamebuf[0] == '/') {
      strlcpy (theFile->linkpath, linknamebuf, PATH_MAX+1);
    } else {
      tmp = sh_util_basename(theFile->fullpath);
      strlcpy (theFile->linkpath, 
	       tmp, 
	       PATH_MAX+1);
      SH_FREE(tmp);
      strlcat (theFile->linkpath, 
	       "/", PATH_MAX+1);
      strlcat (theFile->linkpath, 
	       linknamebuf, 
	       PATH_MAX+1);
    }

    /* stat the link
     */
    stat_return = lstat (theFile->linkpath, &lbuf); 

    /* check for error
     */
    if (stat_return != 0) { 
      if (errno != ENOENT) {
	stat_return = errno;
	tmp2 = sh_util_safe_name (theFile->linkpath);
	sh_error_handle (level, PLACE
			 _("Could not check linked file %s"),
			 _("Could not check linked file %s: %s"),
			 tmp2,
			 sh_error_message (stat_return)
			 );
	theFile->linkisok = BAD;
	SH_FREE(tmp2);
	SH_FREE(linknamebuf);
	return EXIT_FAILURE;
      } else if (errno == ENOENT) {
	theFile->linkisok = BAD;
	stat_return = errno;
	tmp  = sh_util_safe_name (theFile->fullpath);
	tmp2 = sh_util_safe_name (theFile->linkpath);
	sh_error_handle (level, PLACE
			 _("Dangling link %s -> %s"),
			 _("Dangling link %s -> %s: %s"),
			 tmp,
			 tmp2,
			 sh_error_message (stat_return)
			 );
	theFile->linkisok = BAD;
	SH_FREE(tmp);
	SH_FREE(tmp2);
	SH_FREE(linknamebuf);
	return EXIT_FAILURE;
      }
    }

    theFile->linkisok = GOOD;
      

    /* determine file type
     */
    sh_unix_getinfo_type (&lbuf, &type, theFile->link_c_mode);
    theFile->type = type;
  
    /* determine permissions
     */
    sh_unix_getinfo_mode (&lbuf, &mode, theFile->link_c_mode);
    theFile->linkmode = mode;

    /* output the link
     */
    if (theFile->linkisok == GOOD) {
      tmp2 = sh_util_safe_name (linknamebuf);      
      sh_error_handle (SH_ERR_ALL, PLACE
		       NULL,
		       "   >>>  %10s  %s",
		       theFile->link_c_mode,
		       tmp2);
      SH_FREE(tmp2);
    }

    SH_FREE(linknamebuf);

  }
  
  return EXIT_SUCCESS;
}

int sh_unix_is_secure_dir (ShErrLevel level, char * tmp)
{
  char   *    pwd;
  int         error = 0;
  char   *    tmpdir;
  struct stat buf1;

  if (chdir(tmp) != 0) 
      error = errno;

  else {
      pwd = SH_ALLOC(PATH_MAX+1);
      if (NULL == getcwd (pwd, PATH_MAX)) 
	error = errno;
      else if (0 == strncmp(tmp, pwd, strlen(tmp)-1)) {
	if (0 == lstat (tmp, &buf1)) {
	  if ((buf1.st_uid == geteuid()) &&
	      ((buf1.st_mode & S_IWOTH) == 0)) {
	    SH_FREE(pwd);
	    chdir("/");
	    return GOOD;
	  } else {
	      tmpdir  = sh_util_safe_name (tmp);
	      if (buf1.st_uid != geteuid())
		sh_error_handle (level, PLACE
				 NULL,
				 _("Not owner of directory: %s"),
				 tmpdir);
	      if ( (buf1.st_mode & S_IWOTH) != 0)
		sh_error_handle (level, PLACE
				 NULL,
				 _("World-writeable directory: %s"),
				 tmpdir);
	      SH_FREE(tmpdir);
	  }
	} else 
	  error = errno;  /* for lstat */
      } else { /* strncmp failed */
	  tmpdir  = sh_util_safe_name (tmp);
	  sh_error_handle (level, PLACE
			   NULL,
			   _("Symlink in directory path: %s"),
			   tmpdir);
	  SH_FREE(tmpdir);
      }
      SH_FREE(pwd);
  }

  chdir("/");
  if (error != 0) {
    tmpdir  = sh_util_safe_name (tmp);
    sh_error_handle (level, PLACE
		     _("Failed testing security for: %s"),
		     _("Failed testing security for: %s: %s"),
		     tmpdir);
    SH_FREE(tmpdir);
  }
  return BAD;
}

void sh_unix_unlock()
{
  struct stat buf;
  char   *    tmp;
  int         error = 0;
  
  /* logfile is not locked to us
   */
  if (sh.locked == BAD) 
    return;

  /* check whether the directory is secure
   */
  tmp = sh_util_basename(DEFAULT_ERRLOCK);
  if (sh_unix_is_secure_dir (SH_ERR_FATAL, tmp) == BAD) {
    sh_error_handle(SH_ERR_ERR, PLACE
		    _("Error deleting lock file %s"),
		    _("Error deleting lock file %s: not in secure directory"),
		    DEFAULT_ERRLOCK);
    SH_FREE(tmp);
    return;
  }
  SH_FREE(tmp);

  /* delete the lock file if more than one hardlink exists */
  if (0 == lstat (DEFAULT_ERRLOCK, &buf)) {
    if (buf.st_nlink > 1) {
      unlink (DEFAULT_ERRLOCK);
      sh.locked = BAD; /* not locked anymore */
    } else {
      sh_error_handle (SH_ERR_ERR, PLACE
		       _("Error deleting lock file %s"),
		       _("Error deleting lock file %s: st_nlink < 2"),
		       DEFAULT_ERRLOCK);
    }
  } else {
    error = errno;
  }


  if (error != 0 && error != ENOENT) 
    sh_error_handle (SH_ERR_ERR, PLACE
		     _("Error deleting lock file %s"),
		     _("Error deleting lock file %s: %s"),
		     DEFAULT_ERRLOCK,
		     sh_error_message(error));
  
  return;
}

void sh_unix_lock (char * s, char * lock)
{
  struct stat buf;

  if (s == NULL && lock == NULL) return;

  if (0 == lstat (lock, &buf)) {
    fprintf(stderr, _("Startup error: lockfile %s already exists\n"), lock);
    sh.locked       = BAD;
    exit (EXIT_FAILURE);
  }

  link (s, lock);

  if (0 != lstat (lock, &buf)) {
    fprintf(stderr, _("Startup error: error creating lockfile %s\n"), lock);
    sh.locked       = BAD;
    exit (EXIT_FAILURE);
  }

  sh.locked       = GOOD; /* logfile is locked to us */
  return;
}

/***************************************
int sh_unix_test_proc()
{
#ifdef STATFS_WORKS
  struct statfs       buf;
  int                 error;

  if (0 != statfs("/proc", &buf)) {
    error = errno;
    sh_error_handle(SH_ERR_ERR, PLACE
		      _("Cannot access /proc"),
		      _("Failed statfs() on /proc: %s"),
		      sh_error_message(error));
    return BAD;
  } else {
#ifdef SH_PROC_MAGIC
    if (buf.f_type != SH_PROC_MAGIC) {
      sh_error_handle(SH_ERR_ERR, PLACE
		      NULL,
		      _("Not a proc filesystem: /proc"));
      return BAD;
    }
#endif
    return GOOD;
  }
#else
  return GOOD;
#endif
}
************************************************/

/* test whether file exists
 */
int sh_unix_file_exists(char * file_name)
{
  
  struct stat buf;
  if (stat(file_name, &buf) == -1)
    return BAD;
  else
    return GOOD;
}
