/* SAMHAIN file system integrity testing                                   */
/* Copyright (C) 1999 Rainer Wichmann                                      */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "config.h"

#include "samhain.h"
#include "sh_error.h"
#include "sh_utils.h"
#include "sh_unix.h"
#include "sh_tiger.h"
#include "sh_entropy.h"

#ifndef DEFKEY
#define DEFKEY __TIME__
#endif

/* An even simpler en-/decoder, based on Vernam cipher. We use the
 * message as salt to obtain a different one-time pad each time.
 * Should be safe against a listener on the network, but not against someone
 * with read access to the binary.
 */
void sh_util_encode (char * data, char * salt, int mode)
{
  char  cc[16] = "0123456789ABCDEF";
  register int   i, j, j1 = 0, j2 = 0, j3;

  strncpy(skey->vernam, DEFKEY, (int)(0.5*KEY_MAX-1));
  skey->vernam[(int)(0.5*KEY_MAX-1)] = '\0';
  strncat(skey->vernam, salt, (int)(0.5*KEY_MAX-1));
  skey->vernam[KEY_MAX] = '\0';

  strlcpy(skey->vernam, sh_tiger_hash(skey->vernam, TIGER_DATA), KEY_MAX+1);

  /*  fprintf (stderr, "IN : %s\n", data); */
  j = 0;
  while (j < KEY_MAX)
    {
      for (i = 0; i < 16; ++i)
	{
	  if (cc[i] == data[j])   j1 = i;
	  if (cc[i] == skey->vernam[j])    j2 = i;
	}
      if (mode == 0)
	{
	  j3 = j1 + j2;
	  if (j3 > 15) j3 -= 16;
	  data[j] = cc[j3];
	}
      else
	{
	  j3 = j1 - j2;
	  if (j3 <  0) j3 += 16;
	  data[j] = cc[j3];
	}
      if (j == 16 || j == 33) data[j] = ' ';
      ++j;
    }
  /*  fprintf (stderr, "OUT: %s\n", data); */
  return;
}

/* a simple compressor
 */
long sh_util_compress (char * dest, char * src, unsigned long dest_size)
{
  char * add;
  char * get;
  long   count = 0;
  long   dest_end;
  
  dest_end = strlen(dest);
  add      = &dest[dest_end];
  get      = src;

  /********
  fprintf(stderr, "A - COMPRESS IN : %5d %s\n", strlen(src), src);
  fprintf(stderr, "A - COMPRESS OUT: %5d %s\n", strlen(dest), dest);
  fprintf(stderr, "A - count: %5ld   (dest_size-dest_end) %5ld\n", 
	  count, (dest_size-dest_end));
  **********/

  while (count < (dest_size-dest_end))
    {
      if (isalnum(*get)) 
	{
	  *add = *get;
	  ++add;
	  ++count;
	}
      ++get; 
      if (*get == '\0') /* end of src reached */
	{
	  *add = *get;  /* copy the '\0'      */
	  break;        /* and stop copying   */
	}
    }

  /**********
  fprintf(stderr, "B - COMPRESS IN : %5d %s\n", strlen(src), src);
  fprintf(stderr, "B - COMPRESS OUT: %5d %s\n", strlen(dest), dest);
  fprintf(stderr, "B - count: %5ld   (dest_size-dest_end) %5ld\n", 
	  count, (dest_size-dest_end));
  *********/

  dest[dest_size-1] = '\0'; /* paranoia       */
  return count;         /* no of chars copyed */   
}

int sh_util_setlooptime (char * str)
{
  unsigned long i = atoi (str);
  
  if (i >= 0 && i < LONG_MAX) {
    sh.looptime = i;
    return GOOD;
  } else {
    sh_error_handle (SH_ERR_ERR, PLACE
		     NULL,
		     _("Invalid argument for loop time (%s)"),
		     str);
    return BAD;
  }
}

int  sh_util_setchecksum (char * str)
{
  if (strncmp (str, "init", sizeof("init")-1) == 0)
    sh.checkSum = SH_CHECK_INIT;
  else if (strncmp (str, "check", sizeof("check")-1) == 0)
    sh.checkSum = SH_CHECK_CHECK;
  else if (strncmp (str, "none", sizeof("none")-1) == 0)
    sh.checkSum = SH_CHECK_NONE;
  else {
    sh_error_handle (SH_ERR_ERR, PLACE
		     NULL,
		     _("Invalid argument for signature testing (%s),"\
		       "must be 'init', 'check', or 'none'"),
		     str);
    return BAD;
  }
  return GOOD;
}
    

int sh_util_keyinit (char * buf, long size)
{
  char bufx[64+1];

#ifdef HAVE_URANDOM
  FILE * fd2;
  int    count, m_count = 0;

  /* collect at most 24 byte == 192 bits from /dev/random
   */
  if (GOOD == sh_unix_file_exists("/dev/random")) {
    if ( (fd2 = fopen ("/dev/random", "r")) != NULL) {
      while (m_count < ((size < 24) ? size : 24)) {
	count = fread (&buf[m_count], 
		       1, 
		       size-m_count, 
		       fd2 );
	if (count == 0 && ferror(fd2) != 0) break;
	m_count += count;
      }
    }
    buf[size-1] = '\0';
    return 0;
  }
#else
  int fd2;

  fd2 = (int) strlcpy(buf, sh_entropy (), size);
  if (fd2 > 0) 
    return (0);
#endif

  /* simple backup - really bad
   */
  sprintf(bufx, "%ld", sh_unix_longtime ());          /* known to fit  */
  strlcpy(buf, sh_tiger_hash (bufx, TIGER_DATA), size);
  sh_error_handle (SH_ERR_ERR, PLACE
		   NULL,
		   _("Insecure key generation used"));
  return (-1);
}


int sh_util_printf_maxlength (const char * fmt, va_list  vl)
{
  int  length = 1;
  int  fini = 0;
  int  islong, islongdouble;
  char * string_arg;

  while (*fmt) {

    if ( (*fmt) == '%' ) { /* a format specifier */

      fmt++;        /* point to first char after '%' */

      fini = 0;
      islong = 0;
      islongdouble = 0;

      while (*fmt && (fini == 0) ) {
	
	switch (*fmt) {

	case '*':      /* field width supplied by an integer */
	  length = length + va_arg (vl, int);
	  ++fmt;
	  break;
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	  length = length + strtol (fmt, (char**) &fmt, 10);
	  /* strtol makes FastForward to first invalid char */
	  break;

	case 'l':   /* 'long' modifier */
	  islong = 1;
	  ++fmt;
	  break;
	case 'L':  /* 'long double' modifier */ 
	  islongdouble = 1;
	  ++fmt;
	  break;

	case 'd':
	case 'i':
	case 'o':
	case 'u':
	case 'x':
	case 'X':
	  if (islong == 1)
	    (void) va_arg (vl, long);
	  else
	    (void) va_arg (vl, int);
	  ++fmt;
	  fini = 1;
	  break;

	case 'D':
	case 'O':
	case 'U':
	  (void) va_arg (vl, long);
	  length = length + 64;
	  fmt++;
	  fini = 1;
	  break;

	case 'e':
	case 'E':
	case 'f':
	case 'g':
	  if (islongdouble == 1)
	    (void) va_arg (vl, long double);
	  else
	    (void) va_arg (vl, double);
	  length = length + 64;
	  fini = 1;
	  ++fmt;
	  break;

	case 's':
	  string_arg = va_arg (vl, char *);
	  if (string_arg != NULL)
	    length = length + strlen (string_arg);
	  else
	    length = length + 16;
	  fini = 1;
	  ++fmt;
	  break;

	case 'c':
	  (void) va_arg (vl, int);
	  length = length + 1;
	  fini = 1;
	  ++fmt;
	  break;

	case 'p':
	case 'n':
	  (void) va_arg (vl, void * );
	  length = length + 32;
	  fini = 1;
	  ++fmt;
	  break;

	case '%':            /* %% will print '%' */
	  length = length + 1;
	  fini = 1;
	  ++fmt;
	  break;

	default:
	  length = length + 1;
	  ++fmt;
	  break;


	}  /* end switch */
	
      }    

      /* end parsing a single format specifier */

    } else {
      length = length + 1;
      fmt++;
    }

  }

  return length;
}

int sh_util_obscurename (ShErrLevel level, char * name_orig)
{
  char * name = name_orig;
  char * safe;


  /* check name
   */
  while (*name) {
    if ( (*name) == ' '  || (*name) == '"'  || 
	 (*name) == '\b' || (*name) == '\f' || 
	 (*name) == '\n' || (*name) == '\r' || (*name) == '\t' ||
	 (*name) == '\v' || iscntrl(*name)  || !isgraph (*name) ) {

      safe = sh_util_safe_name (name_orig);

      if ( iscntrl(*name) )
	sh_error_handle (level, PLACE
			 NULL,
			 _("File/Dir name with control character: %s"),
			 safe);
      else if ( !isgraph (*name) )
	sh_error_handle (level, PLACE
			 NULL,
			 _("File/Dir name with non-printable charakter: %s"),
			 safe);
      else if ( (*name) ==  '"')
	sh_error_handle (level, PLACE
			 NULL,
			 _("File/Dir name with quote charakter: %s"),
			 safe);
      else
	sh_error_handle (level, PLACE
			 NULL,
			 _("File/Dir name with blank character: %s"),
			 safe);
      if (safe != NULL) SH_FREE(safe);
      return EXIT_FAILURE;
    }
    name++;
  }

  return EXIT_SUCCESS;
}

/* returns freshly allocated memory, return value should be free'd
 */
char * sh_util_basename(char * fullpath)
{
  char * retval;
  int i;

  if (fullpath == NULL) return NULL;

  i = strlen (fullpath);  /* fullpath[i] is terminating '\0' */

  while (i > 0) {
    --i;
    if (fullpath[i] == '/') break;
  }

  if ((fullpath[i] != '/') || (i == 0) ) return NULL;

  retval = SH_ALLOC(i + 1);

  strlcpy (retval, fullpath, i+1);

  return retval;
}

/* returns freshly allocated memory, return value should be free'd
 */
char * sh_util_filename(char * fullpath)
{
  char * retval;
  char * c;
  int i;

  if (fullpath == NULL) return NULL;

  c = strrchr(fullpath, '/');
  i = strlen (c);
  if (i <= 1) return NULL; /* ends in '/' */
  ++c;
  --i;


  retval = SH_ALLOC(i + 1);

  strlcpy (retval, c, i+1);

  return retval;
}

    
/* returns freshly allocated memory, return value should be free'd
 */
char * sh_util_safe_name (char * name)
{
  register int  i = 0;
  char        * p;
  char        * retval;
  char          oct[32];

  retval = SH_ALLOC(4 * strlen(name) + 2); 

  p = name;

  while (*p) {

    if ( (*p) == '\\') {           /* backslash        */
      retval[i] = '\\'; ++i; 
      retval[i] = '\\';
    } else if ( (*p) == '\n') {    /* newline          */
      retval[i] = '\\'; ++i; 
      retval[i] = 'n';
    } else if ( (*p) == '\b') {    /* backspace        */
      retval[i] = '\\'; ++i; 
      retval[i] = 'b';
    } else if ( (*p) == '\r') {    /* carriage  return */
      retval[i] = '\\'; ++i; 
      retval[i] = 'r';
    } else if ( (*p) == '\t') {    /* horizontal tab   */
      retval[i] = '\\'; ++i; 
      retval[i] = 't';
    } else if ( (*p) == '\v') {    /* vertical tab     */
      retval[i] = '\\'; ++i; 
      retval[i] = 'v';
    } else if ( (*p) == '\f') {    /* form-feed        */
      retval[i] = '\\'; ++i; 
      retval[i] = 'f';
    } else if ( (*p) == ' ') {     /* space            */
      retval[i] = '\\'; ++i; 
      retval[i] = ' ';
    } else if ( (*p) == '"') {     /* double quote     */
      retval[i] = '\\'; ++i; 
      retval[i] = '\"';
    } else if (!isgraph (*p)) {    /* not printable    */
      sprintf(oct, "\\%03o", *p);                     /* known to fit  */
      retval[i] = oct[0]; ++i;
      retval[i] = oct[1]; ++i;
      retval[i] = oct[2]; ++i;
      retval[i] = oct[3]; ++i;
    } else {
      retval[i] = *p;
    }
    ++p;
    ++i;
  }

  retval[i] = '\0';

  return retval;
}

int sh_util_isnum (char *str)
{
  char *p = str;
  while (p) {
    if ( ! isdigit(*p) ) return BAD;
    ++p;
  }
  return GOOD;
}

/*********************************************************

char * sh_util_strconcat (const char * arg1, ...)
{
  int     length;
  char    * s;
  char    * strnew;
  va_list vl;

  if (arg1 == NULL) return NULL;

  length = strlen (arg1) + 1;

  va_start (vl, arg1);
  s = va_arg (vl, char * );
  while (s)
    {
      length = length + strlen (s);
      s = va_arg (vl, char * );
    }
  va_end (vl);

  strnew = SH_ALLOC( length + 2 ); 
  strnew[0] = '\0';

  strlcpy (strnew, arg1, length + 2); 

  va_start (vl, arg1);
  s = va_arg (vl, char * );
  while (s)
    {
      strlcat (strnew, s, length + 2);
      s = va_arg (vl, char * );
    }
  va_end (vl);

  return strnew;
}

#include <regex.h>

int sh_util_checkmatch (char * in_str, char * regex_str)
{
  int      status;
  regex_t  preg;
  char *   errbuf;

  status = regcomp(&preg, regex_str, REG_NOSUB);
  if (status != 0) {
    errbuf = SH_ALLOC(BUFSIZ+2);
    regerror(status, &preg, errbuf, BUFSIZ); 
    errbuf[BUFSIZ] = '\0';
    sh_error_handle (SH_ERR_ERR, PLACE
		     _("Problem with regular expression: %s"),
		     _("Failed regcomp() of: %s: %s"),
		     regex_str,
		     errbuf);
    SH_FREE(errbuf);
    return BAD;
  }
	
  if (regexec(&preg, in_str, 0, NULL, 0) == 0) {
    regfree (&preg);
    return GOOD;
  }

  regfree(&preg);
  return BAD;
}

*************************************************************/

#ifndef HAVE_STRLCAT

#include <sys/types.h>
#include <string.h>

/*
 * An alternative implementation of the OpenBSD strlcat() function.
 *
 * Appends src to string dst of size siz (unlike strncat, siz is the
 * full size of dst, not space left).  At most siz-1 characters
 * will be copied.  Always NUL terminates (unless siz == 0).
 * Returns strlen(src); if retval >= siz, truncation occurred.
 */
size_t strlcat(char * dst, const char *src, size_t siz)
{
  register size_t dst_end;
  register size_t dst_free;
  register size_t src_len;

  /* How long is src ?
   */
  src_len  = strlen(src);

  if (siz > 0) {

    /* How much free space do we have ?
     */
    dst_end  = strlen(dst);
    dst_free = siz - dst_end;

    /* Copy at most dst_free chars.
     * The documentation on strncat() is inconclusive as to
     * whether the NULL counts or not, thus use strncpy().
     * dst[dst_end] is the NULL terminator of dst.
     */
    (void) strncpy(&dst[dst_end], 
		   src, 
		   ( (src_len < dst_free) ? src_len : dst_free));

    /* NULL terminate dst.
     */
    if (src_len < dst_free) 
      dst[dst_end+src_len] = '\0';
    dst[siz-1] = '\0';
  }

  return (src_len);
}

#endif /* HAVE_STRLCAT */



#ifndef HAVE_STRLCPY

#include <sys/types.h>
#include <string.h>

/*
 * An alternative implementation of the OpenBSD strlcpy() function.
 *
 * Copy src to string dst of size siz.  At most siz-1 characters
 * will be copied.  Always NUL terminates (unless siz == 0).
 * Returns strlen(src); if retval >= siz, truncation occurred.
 */
size_t strlcpy(char * dst, const char * src, size_t siz)
{
  if (siz > 0) {
    /* copy siz-1 characters 
     */
    strncpy(dst, src, siz-1);

    /* NULL terminate
     */
    dst[siz-1] = '\0';
  }
  return strlen(src);
}
#endif   /* HAVE_STRLCPY */
